"""Test energy save switch rules."""

import collections
import unittest.mock

import HABApp.rule.rule

import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.energy.config.virtual_energy_meter
import habapp_rules.energy.virtual_energy_meter
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer
from habapp_rules.energy.config.virtual_energy_meter import PowerMapping


class TestVirtualEnergyMeterSwitch(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing VirtualEnergyMeterSwitch."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Power_Dimmer")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Energy_Dimmer")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_only_Power")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_only_Energy")

        self._config_max = habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchConfig(
            items=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchItems(monitored_switch="Unittest_Switch", power_output="Unittest_Power_Dimmer", energy_output="Unittest_Energy_Dimmer"),
            parameter=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchParameter(power=100, energy_update_resolution=1),
        )

        self._config_only_power = habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchConfig(
            items=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchItems(monitored_switch="Unittest_Switch", power_output="Unittest_only_Power"),
            parameter=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchParameter(power=42),
        )

        self._config_only_energy = habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchConfig(
            items=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchItems(monitored_switch="Unittest_Switch", energy_output="Unittest_only_Energy"),
            parameter=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterSwitchParameter(power=10_000),
        )

        self._rule_max = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterSwitch(self._config_max)
        self._rule_only_power = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterSwitch(self._config_only_power)
        self._rule_only_energy = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterSwitch(self._config_only_energy)

    def test_init_with_switch_on(self) -> None:
        """Test init with switch on."""
        tests.helper.oh_item.item_state_change_event("Unittest_Switch", "ON")

        with unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.soon") as run_soon_mock:
            rule = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterSwitch(self._config_max)

        run_soon_mock.assert_called_once_with(rule._cb_monitored_item, unittest.mock.ANY)
        self.assertEqual("Unittest_Switch", run_soon_mock.call_args[0][1].name)
        self.assertEqual("ON", run_soon_mock.call_args[0][1].value)

    def test_get_power(self) -> None:
        """Test _get_power."""
        self.assertEqual(100, self._rule_max._get_power())
        self.assertEqual(42, self._rule_only_power._get_power())
        self.assertEqual(10_000, self._rule_only_energy._get_power())

    def test_get_energy_countdown_time(self) -> None:
        """Test _get_energy_countdown_time."""
        TestCase = collections.namedtuple("TestCase", "power, energy_update_resolution, expected_time")

        test_cases = [
            TestCase(power=10, energy_update_resolution=0.010, expected_time=3600),
            TestCase(power=10, energy_update_resolution=0.001, expected_time=360),
            TestCase(power=100, energy_update_resolution=0.010, expected_time=360),
            TestCase(power=100, energy_update_resolution=0.001, expected_time=36),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self._rule_max._config.parameter.energy_update_resolution = test_case.energy_update_resolution
                self._rule_max._power = test_case.power
                self.assertEqual(test_case.expected_time, self._rule_max._get_energy_countdown_time())

    def test_power(self) -> None:
        """Test if power ist set correctly."""
        tests.helper.oh_item.assert_value("Unittest_Power_Dimmer", 0)
        tests.helper.oh_item.assert_value("Unittest_only_Power", 0)

        tests.helper.oh_item.item_state_change_event("Unittest_Switch", "ON")
        tests.helper.oh_item.assert_value("Unittest_Power_Dimmer", 100)
        tests.helper.oh_item.assert_value("Unittest_only_Power", 42)

        tests.helper.oh_item.item_state_change_event("Unittest_Switch", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Power_Dimmer", 0)
        tests.helper.oh_item.assert_value("Unittest_only_Power", 0)

    def test_energy(self) -> None:
        """Test if energy ist set correctly."""
        tests.helper.oh_item.assert_value("Unittest_Energy_Dimmer", 0)
        tests.helper.oh_item.assert_value("Unittest_only_Energy", 0)

        tests.helper.oh_item.item_state_change_event("Unittest_Switch", "ON")
        tests.helper.oh_item.assert_value("Unittest_Energy_Dimmer", 0)
        tests.helper.oh_item.assert_value("Unittest_only_Energy", 0)

    def test_cb_monitored_item(self) -> None:
        """Test _cb_monitored_item."""
        with (
            unittest.mock.patch.object(self._rule_max, "_reset_countdown") as max_reset_countdown_mock,
            unittest.mock.patch.object(self._rule_max, "_set_energy_from_remaining_time") as max_set_energy_mock,
            unittest.mock.patch.object(self._rule_only_power, "_reset_countdown") as only_reset_countdown_mock,
            unittest.mock.patch.object(self._rule_only_power, "_set_energy_from_remaining_time") as only_set_energy_mock,
        ):
            # ON
            tests.helper.oh_item.item_state_change_event("Unittest_Switch", "ON")
            max_reset_countdown_mock.assert_called_once_with()
            max_set_energy_mock.assert_not_called()
            only_reset_countdown_mock.assert_not_called()
            only_set_energy_mock.assert_not_called()

            # OFF
            max_reset_countdown_mock.reset_mock()
            tests.helper.oh_item.item_state_change_event("Unittest_Switch", "OFF")
            max_reset_countdown_mock.assert_not_called()
            max_set_energy_mock.assert_called_once_with()
            only_reset_countdown_mock.assert_not_called()
            only_set_energy_mock.assert_not_called()

    def test_reset_countdown(self) -> None:
        """Test _reset_countdown."""
        with unittest.mock.patch.object(self._rule_max, "_send_energy_countdown") as countdown_mock, unittest.mock.patch.object(self._rule_max, "_get_energy_countdown_time") as get_time_mock, unittest.mock.patch("time.time", return_value=42):
            self._rule_max._reset_countdown()
            countdown_mock.set_countdown.assert_called_once_with(get_time_mock.return_value)
            countdown_mock.reset.assert_called_once()
            self.assertEqual(42, self._rule_max._last_energy_countdown_reset)

    def test_cb_countdown_end(self) -> None:
        """Test _cb_countdown_end."""
        with (
            unittest.mock.patch.object(self._rule_max, "_update_energy_item") as update_energy_item_mock,
            unittest.mock.patch.object(self._rule_max, "_get_energy_countdown_time") as get_countdown_mock,
            unittest.mock.patch.object(self._rule_max, "_reset_countdown") as reset_countdown_mock,
        ):
            self._rule_max._cb_countdown_end()
            update_energy_item_mock.assert_called_once_with(get_countdown_mock.return_value)
            reset_countdown_mock.assert_called_once()

    def test_update_energy_item(self) -> None:
        """Test _update_energy_item."""
        TestCase = collections.namedtuple("TestCase", "energy_output_value, power, time_since_last_update, expected_energy")

        test_cases = [
            TestCase(energy_output_value=0, power=0, time_since_last_update=0, expected_energy=0),
            TestCase(energy_output_value=0, power=0, time_since_last_update=10, expected_energy=0),
            TestCase(energy_output_value=0, power=100, time_since_last_update=0, expected_energy=0),
            TestCase(energy_output_value=0, power=100, time_since_last_update=10, expected_energy=0.1 * 10 / 3600),
            TestCase(energy_output_value=0, power=-100, time_since_last_update=0, expected_energy=0),
            TestCase(energy_output_value=0, power=-100, time_since_last_update=10, expected_energy=-0.1 * 10 / 3600),
            # with initial power
            TestCase(energy_output_value=1000, power=0, time_since_last_update=0, expected_energy=1000),
            TestCase(energy_output_value=1000, power=0, time_since_last_update=10, expected_energy=1000),
            TestCase(energy_output_value=1000, power=100, time_since_last_update=0, expected_energy=1000),
            TestCase(energy_output_value=1000, power=100, time_since_last_update=10, expected_energy=1000 + 0.1 * 10 / 3600),
            TestCase(energy_output_value=1000, power=-100, time_since_last_update=0, expected_energy=1000),
            TestCase(energy_output_value=1000, power=-100, time_since_last_update=10, expected_energy=1000 - 0.1 * 10 / 3600),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self._rule_max._config.items.energy_output.value = test_case.energy_output_value
                self._rule_max._power = test_case.power
                self._rule_max._update_energy_item(test_case.time_since_last_update)
                self.assertEqual(test_case.expected_energy, self._rule_max._config.items.energy_output.value)


class TestVirtualEnergyMeterNumber(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing VirtualEnergyMeterNumber."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Power_Dimmer")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Energy_Dimmer")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Power_Number")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Energy_Number")

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_only_Power")
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_only_Energy")

        self._config_max_dimmer = habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberConfig(
            items=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberItems(monitored_item="Unittest_Dimmer", power_output="Unittest_Power_Dimmer", energy_output="Unittest_Energy_Dimmer"),
            parameter=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberParameter(power_mapping=[PowerMapping(0, 0), PowerMapping(100, 1000)], energy_update_resolution=1),
        )

        self._config_max_number = habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberConfig(
            items=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberItems(monitored_item="Unittest_Number", power_output="Unittest_Power_Number", energy_output="Unittest_Energy_Number"),
            parameter=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberParameter(power_mapping=[PowerMapping(0, 0), PowerMapping(2, 100)], energy_update_resolution=1),
        )

        self._config_only_power = habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberConfig(
            items=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberItems(monitored_item="Unittest_Dimmer", power_output="Unittest_only_Power"),
            parameter=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberParameter(power_mapping=[PowerMapping(0, 0), PowerMapping(100, 42)]),
        )

        self._config_only_energy = habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberConfig(
            items=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberItems(monitored_item="Unittest_Dimmer", energy_output="Unittest_only_Energy"),
            parameter=habapp_rules.energy.config.virtual_energy_meter.EnergyMeterNumberParameter(power_mapping=[PowerMapping(0, 0), PowerMapping(100, 10_000)]),
        )

        self._rule_max_dimmer = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterNumber(self._config_max_dimmer)
        self._rule_max_number = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterNumber(self._config_max_number)
        self._rule_only_power = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterNumber(self._config_only_power)
        self._rule_only_energy = habapp_rules.energy.virtual_energy_meter.VirtualEnergyMeterNumber(self._config_only_energy)

    def test_power(self) -> None:
        """Test if power ist set correctly."""
        tests.helper.oh_item.assert_value("Unittest_Power_Dimmer", 0)
        tests.helper.oh_item.assert_value("Unittest_only_Power", 0)

        tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 100)
        tests.helper.oh_item.assert_value("Unittest_Power_Dimmer", 1000)
        tests.helper.oh_item.assert_value("Unittest_only_Power", 42)

        tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 50)
        tests.helper.oh_item.assert_value("Unittest_Power_Dimmer", 500)
        tests.helper.oh_item.assert_value("Unittest_only_Power", 21)

        tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 0)
        tests.helper.oh_item.assert_value("Unittest_Power_Dimmer", 0)
        tests.helper.oh_item.assert_value("Unittest_only_Power", 0)

    def test_get_energy_countdown_time(self) -> None:
        """Test _get_energy_countdown_time."""
        # power is not yet set -> avoid division by zero and return 1
        self.assertEqual(0, self._rule_max_dimmer._power)
        self.assertEqual(1, self._rule_max_dimmer._get_energy_countdown_time())

    def test_cb_monitored_item(self) -> None:
        """Test _cb_monitored_item."""
        with (
            unittest.mock.patch.object(self._rule_max_dimmer, "_reset_countdown") as max_reset_countdown_mock,
            unittest.mock.patch.object(self._rule_max_dimmer, "_set_energy_from_remaining_time") as max_set_energy_mock,
            unittest.mock.patch.object(self._rule_only_power, "_reset_countdown") as only_reset_countdown_mock,
            unittest.mock.patch.object(self._rule_only_power, "_set_energy_from_remaining_time") as only_set_energy_mock,
        ):
            # ON
            tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 20, 0)
            max_reset_countdown_mock.assert_called_once_with()
            max_set_energy_mock.assert_not_called()
            only_reset_countdown_mock.assert_not_called()
            only_set_energy_mock.assert_not_called()

            # Change
            max_reset_countdown_mock.reset_mock()
            tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 40, 0)
            max_reset_countdown_mock.assert_called_once_with()
            max_set_energy_mock.assert_called_once_with()
            only_reset_countdown_mock.assert_not_called()
            only_set_energy_mock.assert_not_called()

            # OFF
            max_reset_countdown_mock.reset_mock()
            max_set_energy_mock.reset_mock()
            tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 0, 20)
            max_reset_countdown_mock.assert_not_called()
            max_set_energy_mock.assert_called_once_with()
            only_reset_countdown_mock.assert_not_called()
            only_set_energy_mock.assert_not_called()



"""Test ventilation rules."""

import collections
import datetime
import pathlib
import sys
import unittest
import unittest.mock

import HABApp.rule.rule

import habapp_rules.actors.config.shading
import habapp_rules.actors.config.ventilation
import habapp_rules.actors.light
import habapp_rules.actors.shading
import habapp_rules.actors.state_observer
import habapp_rules.actors.ventilation
import habapp_rules.core.exceptions
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system
import tests.helper.graph_machines
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer


class TestGlobalFunctions(unittest.TestCase):
    """Test global functions."""

    def test_to_datetime(self) -> None:
        """Test _to_datetime."""
        TestCase = collections.namedtuple("TestCase", "input_time, now, expected_result")

        test_cases = [
            TestCase(datetime.time(0, 0), datetime.datetime(2024, 1, 1, 0, 1), datetime.datetime(2024, 1, 2, 0, 0)),
            TestCase(datetime.time(6, 0), datetime.datetime(2024, 1, 1, 0, 1), datetime.datetime(2024, 1, 1, 6, 0)),
            TestCase(datetime.time(18, 0), datetime.datetime(2024, 1, 1, 17, 59), datetime.datetime(2024, 1, 1, 18, 0)),
            TestCase(datetime.time(18, 0), datetime.datetime(2024, 1, 1, 18, 00), datetime.datetime(2024, 1, 2, 18, 0)),
            TestCase(datetime.time(18, 0), datetime.datetime(2024, 1, 1, 18, 1), datetime.datetime(2024, 1, 2, 18, 0)),
        ]

        combine_orig = datetime.datetime.combine

        with unittest.mock.patch("datetime.datetime") as datetime_mock:
            datetime.datetime.combine = combine_orig
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    datetime_mock.now.return_value = test_case.now
                    self.assertEqual(test_case.expected_result, habapp_rules.actors.ventilation._to_datetime(test_case.input_time))


class TestVentilation(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing Ventilation."""

    def setUp(self) -> None:
        """Setup test case."""
        self.run_at_mock_patcher = unittest.mock.patch("HABApp.rule.scheduler.job_builder.HABAppJobBuilder.once")
        self.addCleanup(self.run_at_mock_patcher.stop)
        self.run_at_mock = self.run_at_mock_patcher.start()

        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Ventilation_min_level", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_min_manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Ventilation_min_level_state", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Ventilation_max_level", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Ventilation_max_Custom_State", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_hand_request", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_external_request", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_feedback_on", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_feedback_power", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Ventilation_max_display_text", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Presence_state", None)

        parameter_max = habapp_rules.actors.config.ventilation.VentilationParameter(
            state_normal=habapp_rules.actors.config.ventilation.StateConfig(level=101, display_text="Normal Custom"),
            state_hand=habapp_rules.actors.config.ventilation.StateConfigWithTimeout(level=102, display_text="Hand Custom", timeout=42 * 60),
            state_external=habapp_rules.actors.config.ventilation.StateConfig(level=103, display_text="External Custom"),
            state_humidity=habapp_rules.actors.config.ventilation.StateConfig(level=104, display_text="Humidity Custom"),
            state_long_absence=habapp_rules.actors.config.ventilation.StateConfigLongAbsence(level=105, display_text="Absence Custom", duration=1800, start_time=datetime.time(18)),
        )

        config_max = habapp_rules.actors.config.ventilation.VentilationConfig(
            items=habapp_rules.actors.config.ventilation.VentilationItems(
                ventilation_level="Unittest_Ventilation_max_level",
                manual="Unittest_Ventilation_max_manual",
                hand_request="Unittest_Ventilation_max_hand_request",
                external_request="Unittest_Ventilation_max_external_request",
                feedback_on="Unittest_Ventilation_max_feedback_on",
                feedback_power="Unittest_Ventilation_max_feedback_power",
                display_text="Unittest_Ventilation_max_display_text",
                presence_state="Unittest_Presence_state",
                state="Unittest_Ventilation_max_Custom_State",
            ),
            parameter=parameter_max,
        )

        config_min = habapp_rules.actors.config.ventilation.VentilationConfig(
            items=habapp_rules.actors.config.ventilation.VentilationItems(ventilation_level="Unittest_Ventilation_min_level", manual="Unittest_Ventilation_min_manual", state="H_Unittest_Ventilation_min_level_state"),
        )

        self.ventilation_min = habapp_rules.actors.ventilation.Ventilation(config_min)
        self.ventilation_max = habapp_rules.actors.ventilation.Ventilation(config_max)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "Ventilation"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(
            model=tests.helper.graph_machines.FakeModel(), states=self.ventilation_min.states, transitions=self.ventilation_min.trans, initial=self.ventilation_min.state, show_conditions=False
        )

        graph.get_graph().draw(picture_dir / "Ventilation.png", format="png", prog="dot")

        for state_name in [state for state in self._get_state_names(self.ventilation_min.states) if "init" not in state.lower()]:
            graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self.ventilation_min.states, transitions=self.ventilation_min.trans, initial=state_name, show_conditions=True)
            graph.get_graph(force_new=True, show_roi=True).draw(picture_dir / f"Ventilation_{state_name}.png", format="png", prog="dot")

    def test_init(self) -> None:
        """Test __init__."""
        # check timeouts
        self.assertEqual(3600, self.ventilation_min.state_machine.get_state("Auto_PowerHand").timeout)
        self.assertEqual(3600, self.ventilation_min.state_machine.get_state("Auto_LongAbsence_On").timeout)

        self.assertEqual(42 * 60, self.ventilation_max.state_machine.get_state("Auto_PowerHand").timeout)
        self.assertEqual(1800, self.ventilation_max.state_machine.get_state("Auto_LongAbsence_On").timeout)

    def test_get_initial_state(self) -> None:
        """Test getting initial state."""
        TestCase = collections.namedtuple("TestCase", "presence_state, manual, hand_request, external_request, expected_state_min, expected_state_max")

        test_cases = [
            # present
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, False, False, False, "Auto_Normal", "Auto_Normal"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, False, False, True, "Auto_Normal", "Auto_PowerExternal"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, False, True, False, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, False, True, True, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, True, False, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, True, False, True, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, True, True, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, True, True, True, "Manual", "Manual"),
            # long absence
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, False, False, False, "Auto_Normal", "Auto_LongAbsence"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, False, False, True, "Auto_Normal", "Auto_LongAbsence"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, False, True, False, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, False, True, True, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, True, False, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, True, False, True, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, True, True, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, True, True, True, "Manual", "Manual"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Ventilation_min_manual", "ON" if test_case.manual else "OFF")
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_manual", "ON" if test_case.manual else "OFF")
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_hand_request", "ON" if test_case.hand_request else "OFF")
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_external_request", "ON" if test_case.external_request else "OFF")
                tests.helper.oh_item.set_state("Unittest_Presence_state", test_case.presence_state)

                self.assertEqual(test_case.expected_state_min, self.ventilation_min._get_initial_state())
                self.assertEqual(test_case.expected_state_max, self.ventilation_max._get_initial_state())

    def test_set_level(self) -> None:
        """Test _set_level."""
        TestCase = collections.namedtuple("TestCase", "state, expected_level")

        test_cases = [
            TestCase("Manual", None),
            TestCase("Auto_PowerHand", 102),
            TestCase("Auto_Normal", 101),
            TestCase("Auto_PowerExternal", 103),
            TestCase("Auto_LongAbsence_On", 105),
            TestCase("Auto_LongAbsence_Off", 0),
            TestCase("Auto_Init", None),
        ]

        with unittest.mock.patch("habapp_rules.core.helper.send_if_different") as send_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    send_mock.reset_mock()
                    self.ventilation_max.state = test_case.state

                    self.ventilation_max._set_level()

                    if test_case.expected_level is not None:
                        send_mock.assert_called_once_with(self.ventilation_max._config.items.ventilation_level, test_case.expected_level)
                    else:
                        send_mock.assert_not_called()

    def test_set_feedback_states(self) -> None:
        """Test _set_feedback_states."""
        TestCase = collections.namedtuple("TestCase", "ventilation_level, state, expected_on, expected_power, expected_display_text")

        test_cases = [
            TestCase(None, "Auto_PowerHand", False, False, "Hand Custom 42min"),
            TestCase(None, "Auto_Normal", False, False, "Normal Custom"),
            TestCase(None, "Auto_PowerExternal", False, False, "External Custom"),
            TestCase(None, "Auto_LongAbsence_On", False, False, "Absence Custom ON"),
            TestCase(None, "Auto_LongAbsence_Off", False, False, "Absence Custom OFF"),
            TestCase(None, "Auto_Init", False, False, "Absence Custom OFF"),
            TestCase(0, "Auto_PowerHand", False, False, "Hand Custom 42min"),
            TestCase(0, "Auto_Normal", False, False, "Normal Custom"),
            TestCase(0, "Auto_PowerExternal", False, False, "External Custom"),
            TestCase(0, "Auto_LongAbsence_On", False, False, "Absence Custom ON"),
            TestCase(0, "Auto_LongAbsence_Off", False, False, "Absence Custom OFF"),
            TestCase(0, "Auto_Init", False, False, "Absence Custom OFF"),
            TestCase(1, "Auto_PowerHand", True, False, "Hand Custom 42min"),
            TestCase(1, "Auto_Normal", True, False, "Normal Custom"),
            TestCase(1, "Auto_PowerExternal", True, False, "External Custom"),
            TestCase(1, "Auto_LongAbsence_On", True, False, "Absence Custom ON"),
            TestCase(1, "Auto_LongAbsence_Off", True, False, "Absence Custom OFF"),
            TestCase(1, "Auto_Init", True, False, "Absence Custom OFF"),
            TestCase(2, "Auto_PowerHand", True, True, "Hand Custom 42min"),
            TestCase(2, "Auto_Normal", True, True, "Normal Custom"),
            TestCase(2, "Auto_PowerExternal", True, True, "External Custom"),
            TestCase(2, "Auto_LongAbsence_On", True, True, "Absence Custom ON"),
            TestCase(2, "Auto_LongAbsence_Off", True, True, "Absence Custom OFF"),
            TestCase(2, "Auto_Init", True, True, "Absence Custom OFF"),
            TestCase(42, "Auto_PowerHand", True, True, "Hand Custom 42min"),
            TestCase(42, "Auto_Normal", True, True, "Normal Custom"),
            TestCase(42, "Auto_PowerExternal", True, True, "External Custom"),
            TestCase(42, "Auto_LongAbsence_On", True, True, "Absence Custom ON"),
            TestCase(42, "Auto_LongAbsence_Off", True, True, "Absence Custom OFF"),
            TestCase(42, "Auto_Init", True, True, "Absence Custom OFF"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.ventilation_min._ventilation_level = test_case.ventilation_level
                self.ventilation_max._ventilation_level = test_case.ventilation_level
                self.ventilation_min.state = test_case.state
                self.ventilation_max.state = test_case.state

                self.ventilation_min._set_feedback_states()
                self.ventilation_max._set_feedback_states()

                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_feedback_on", "ON" if test_case.expected_on else "OFF")
                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_feedback_power", "ON" if test_case.expected_power else "OFF")
                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_display_text", test_case.expected_display_text)

    def test_on_enter_long_absence_off(self) -> None:
        """Test on_enter_Auto_LongAbsence_Off."""
        with unittest.mock.patch.object(self.ventilation_max, "_trigger_long_absence_power_on") as trigger_on_mock, unittest.mock.patch("habapp_rules.actors.ventilation._to_datetime") as to_datetime_mock:
            self.ventilation_max.to_Auto_LongAbsence_Off()
        self.run_at_mock.assert_called_once_with(to_datetime_mock.return_value, trigger_on_mock)

    def test_trigger_long_absence_power_on(self) -> None:
        """Test _trigger_long_absence_power_on."""
        with unittest.mock.patch.object(self.ventilation_max, "_long_absence_power_on") as power_on_mock:
            self.ventilation_max._trigger_long_absence_power_on()
        power_on_mock.assert_called_once()

    def test__set_hand_display_text(self) -> None:
        """Test __set_hand_display_text."""
        # wrong state
        for state in ["Manual", "Auto_Normal", "Auto_PowerHumidity", "Auto_PowerDryer", "Auto_LongAbsence_On", "Auto_LongAbsence_Off"]:
            self.ventilation_max.state = state

            self.ventilation_max._VentilationBase__set_hand_display_text()
            self.run_at_mock.assert_not_called()

        # PowerHand state:
        TestCase = collections.namedtuple("TestCase", "changed_time, now_time, expected_display")

        test_cases = [
            TestCase(datetime.datetime(2024, 1, 1, 12), datetime.datetime(2024, 1, 1, 12, 0), "Hand Custom 42min"),
            TestCase(datetime.datetime(2024, 1, 1, 12), datetime.datetime(2024, 1, 1, 12, 0, 30), "Hand Custom 42min"),
            TestCase(datetime.datetime(2024, 1, 1, 12), datetime.datetime(2024, 1, 1, 12, 0, 31), "Hand Custom 41min"),
            TestCase(datetime.datetime(2024, 1, 1, 12), datetime.datetime(2024, 1, 1, 12, 2, 0), "Hand Custom 40min"),
            TestCase(datetime.datetime(2024, 1, 1, 12), datetime.datetime(2024, 1, 1, 12, 2, 31), "Hand Custom 39min"),
            TestCase(datetime.datetime(2024, 1, 1, 12), datetime.datetime(2024, 1, 1, 12, 42, 0), "Hand Custom 0min"),
            TestCase(datetime.datetime(2024, 1, 1, 12), datetime.datetime(2024, 1, 1, 12, 45, 0), "Hand Custom 0min"),
        ]

        self.ventilation_max.state = "Auto_PowerHand"

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.ventilation_max._state_change_time = test_case.changed_time
                now_value = test_case.now_time
                self.run_at_mock.reset_mock()
                with unittest.mock.patch("datetime.datetime") as datetime_mock:
                    datetime_mock.now.return_value = now_value
                    self.ventilation_max._VentilationBase__set_hand_display_text()
                self.run_at_mock.assert_called_once()
                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_display_text", test_case.expected_display)

    def test_external_active_and_configured(self) -> None:
        """Test _external_active_and_configured."""
        self.assertFalse(self.ventilation_min._external_active_and_configured())

        tests.helper.oh_item.set_state("Unittest_Ventilation_max_external_request", "OFF")
        self.assertFalse(self.ventilation_max._external_active_and_configured())

        tests.helper.oh_item.set_state("Unittest_Ventilation_max_external_request", "ON")
        self.assertTrue(self.ventilation_max._external_active_and_configured())

    def test_auto_normal_transitions(self) -> None:
        """Test transitions of state Auto_Normal."""
        # to Auto_PowerHand
        self.ventilation_min.to_Auto_Normal()
        self.ventilation_max.to_Auto_Normal()

        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "ON")

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)

        # back to Auto_Normal
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "OFF")

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

        # to Auto_PowerExternal
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "ON")

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_PowerExternal", self.ventilation_max.state)

        # back to Auto_Normal
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "OFF")

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

        # to Auto_LongAbsence
        tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", habapp_rules.system.PresenceState.LONG_ABSENCE.value)

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_LongAbsence_Off", self.ventilation_max.state)

        # back to Auto_Normal
        tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", habapp_rules.system.PresenceState.PRESENCE.value)

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

    def test_auto_power_external_transitions(self) -> None:
        """Test transitions of state Auto_PowerExternal."""
        # to Auto_PowerExternal
        self.ventilation_max.to_Auto_PowerExternal()
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "OFF")
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

        # back to AutoPowerExternal
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "ON")
        self.assertEqual("Auto_PowerExternal", self.ventilation_max.state)

        # to Auto_PowerHand
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "ON")
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "ON")
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)

        # back to AutoPowerExternal
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "OFF")
        self.assertEqual("Auto_PowerExternal", self.ventilation_max.state)

        # to Auto_LongAbsence
        tests.helper.oh_item.item_state_change_event("Unittest_Presence_state", habapp_rules.system.PresenceState.LONG_ABSENCE.value)
        self.assertEqual("Auto_LongAbsence_Off", self.ventilation_max.state)

    def test_auto_power_hand_transitions(self) -> None:
        """Test transitions of state Auto_PowerHand."""
        # set Auto_LongAbsence as initial state
        self.ventilation_max.to_Auto_LongAbsence_On()

        # to Auto_PowerHand
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "ON")
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)

        # to Auto_Normal (external request is not ON)
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "OFF")
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

        # back to Auto_PowerHand
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "ON")
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)

        # to Auto_PowerExternal
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "ON")
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "OFF")
        self.assertEqual("Auto_PowerExternal", self.ventilation_max.state)

        # back to Auto_PowerHand
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "ON")
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)

        # timeout
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "OFF")
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_hand_request", "OFF")

    def test_manual_transitions(self) -> None:
        """Test transitions of state Manual."""
        # set Auto as initial state
        self.ventilation_min.to_Auto_Normal()
        self.ventilation_max.to_Auto_Normal()

        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_min_manual", "ON")
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_manual", "ON")

        self.assertEqual("Manual", self.ventilation_min.state)
        self.assertEqual("Manual", self.ventilation_max.state)

        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_min_manual", "OFF")
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_manual", "OFF")

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)


class TestVentilationHeliosTwoStage(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing VentilationHeliosTwoStage."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_min_output_on", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_min_output_power", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_min_manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Ventilation_min_output_on_state", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_output_on", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_output_power", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Ventilation_max_Custom_State", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_hand_request", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_external_request", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_feedback_on", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_feedback_power", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Ventilation_max_feedback_ventilation_level", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Ventilation_max_display_text", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Presence_state", None)

        parameter_max = habapp_rules.actors.config.ventilation.VentilationTwoStageParameter(
            state_normal=habapp_rules.actors.config.ventilation.StateConfig(level=101, display_text="Normal Custom"),
            state_hand=habapp_rules.actors.config.ventilation.StateConfigWithTimeout(level=102, display_text="Hand Custom", timeout=42 * 60),
            state_external=habapp_rules.actors.config.ventilation.StateConfig(level=103, display_text="External Custom"),
            state_humidity=habapp_rules.actors.config.ventilation.StateConfig(level=104, display_text="Humidity Custom"),
            state_long_absence=habapp_rules.actors.config.ventilation.StateConfigLongAbsence(level=105, display_text="Absence Custom", duration=1800, start_time=datetime.time(18)),
            state_after_run=habapp_rules.actors.config.ventilation.StateConfig(level=99, display_text="AfterRun Custom"),
            after_run_timeout=350,
        )

        config_max = habapp_rules.actors.config.ventilation.VentilationTwoStageConfig(
            items=habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
                ventilation_output_on="Unittest_Ventilation_max_output_on",
                ventilation_output_power="Unittest_Ventilation_max_output_power",
                manual="Unittest_Ventilation_max_manual",
                hand_request="Unittest_Ventilation_max_hand_request",
                external_request="Unittest_Ventilation_max_external_request",
                feedback_on="Unittest_Ventilation_max_feedback_on",
                feedback_power="Unittest_Ventilation_max_feedback_power",
                display_text="Unittest_Ventilation_max_display_text",
                presence_state="Unittest_Presence_state",
                state="Unittest_Ventilation_max_Custom_State",
                feedback_ventilation_level="Unittest_Ventilation_max_feedback_ventilation_level",
            ),
            parameter=parameter_max,
        )

        config_min = habapp_rules.actors.config.ventilation.VentilationTwoStageConfig(
            items=habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
                ventilation_output_on="Unittest_Ventilation_min_output_on", ventilation_output_power="Unittest_Ventilation_min_output_power", manual="Unittest_Ventilation_min_manual", state="H_Unittest_Ventilation_min_output_on_state"
            )
        )

        self.ventilation_min = habapp_rules.actors.ventilation.VentilationHeliosTwoStage(config_min)
        self.ventilation_max = habapp_rules.actors.ventilation.VentilationHeliosTwoStage(config_max)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "VentilationHeliosTwoStage"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(
            model=tests.helper.graph_machines.FakeModel(), states=self.ventilation_min.states, transitions=self.ventilation_min.trans, initial=self.ventilation_min.state, show_conditions=False
        )

        graph.get_graph().draw(picture_dir / "Ventilation.png", format="png", prog="dot")

        for state_name in [state for state in self._get_state_names(self.ventilation_min.states) if "init" not in state.lower()]:
            graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self.ventilation_min.states, transitions=self.ventilation_min.trans, initial=state_name, show_conditions=True)
            graph.get_graph(force_new=True, show_roi=True).draw(picture_dir / f"Ventilation_{state_name}.png", format="png", prog="dot")

    def test_set_level(self) -> None:
        """Test _set_level."""
        TestCase = collections.namedtuple("TestCase", "state, expected_on, expected_power, expected_level")

        test_cases = [
            TestCase("Manual", None, None, 101),
            TestCase("Auto_PowerHand", "ON", "ON", 102),
            TestCase("Auto_Normal", "ON", "OFF", 1),
            TestCase("Auto_PowerExternal", "ON", "ON", 103),
            TestCase("Auto_LongAbsence_On", "ON", "ON", 105),
            TestCase("Auto_LongAbsence_Off", "OFF", "OFF", 0),
            TestCase("Auto_Init", None, None, 0),
            TestCase("Auto_PowerAfterRun", "ON", "OFF", 99),
        ]

        self.ventilation_max._config.parameter.state_normal.level = 1

        with unittest.mock.patch("habapp_rules.core.helper.send_if_different") as send_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    send_mock.reset_mock()
                    self.ventilation_max.state = test_case.state

                    self.ventilation_max._set_level()

                    if test_case.expected_on is not None:
                        send_mock.assert_any_call(self.ventilation_max._config.items.ventilation_output_on, test_case.expected_on)

                    if test_case.expected_power is not None:
                        send_mock.assert_any_call(self.ventilation_max._config.items.ventilation_output_power, test_case.expected_power)

                    tests.helper.oh_item.assert_value("Unittest_Ventilation_max_feedback_ventilation_level", test_case.expected_level)

    def test_set_feedback_states(self) -> None:
        """Test _set_feedback_states."""
        TestCase = collections.namedtuple("TestCase", "ventilation_level, state, expected_on, expected_power, expected_display_text")

        test_cases = [TestCase(None, "Auto_PowerAfterRun", False, False, "AfterRun Custom"), TestCase(0, "Auto_PowerAfterRun", False, False, "AfterRun Custom"), TestCase(1, "Auto_PowerAfterRun", True, False, "AfterRun Custom")]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.ventilation_min._ventilation_level = test_case.ventilation_level
                self.ventilation_max._ventilation_level = test_case.ventilation_level
                self.ventilation_min.state = test_case.state
                self.ventilation_max.state = test_case.state

                self.ventilation_min._set_feedback_states()
                self.ventilation_max._set_feedback_states()

                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_feedback_on", "ON" if test_case.expected_on else "OFF")
                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_feedback_power", "ON" if test_case.expected_power else "OFF")
                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_display_text", test_case.expected_display_text)

    def test_power_after_run_transitions(self) -> None:
        """Test transitions of PowerAfterRun."""
        # PowerAfterRun to PowerHand
        self.ventilation_max.to_Auto_PowerAfterRun()
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "ON")
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)

        # back to PowerAfterRun
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "OFF")
        self.assertEqual("Auto_PowerAfterRun", self.ventilation_max.state)

        # PowerAfterRun to PowerExternal
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "ON")
        self.assertEqual("Auto_PowerExternal", self.ventilation_max.state)

        # back to PowerAfterRun
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "OFF")
        self.assertEqual("Auto_PowerAfterRun", self.ventilation_max.state)

        # timeout of PowerAfterRun
        tests.helper.timer.call_timeout(self.transitions_timer_mock)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)


class TestVentilationHeliosTwoStageHumidity(tests.helper.test_case_base.TestCaseBaseStateMachine):
    """Tests cases for testing VentilationHeliosTwoStageHumidity."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBaseStateMachine.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_min_output_on", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_min_output_power", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Ventilation_min_current", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_min_manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "H_Unittest_Ventilation_min_output_on_state", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_output_on", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_output_power", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Ventilation_max_current", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_manual", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Ventilation_max_Custom_State", None)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_hand_request", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_external_request", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_feedback_on", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Ventilation_max_feedback_power", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Ventilation_max_display_text", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_Presence_state", None)

        parameter_max = habapp_rules.actors.config.ventilation.VentilationTwoStageParameter(
            state_normal=habapp_rules.actors.config.ventilation.StateConfig(level=101, display_text="Normal Custom"),
            state_hand=habapp_rules.actors.config.ventilation.StateConfigWithTimeout(level=102, display_text="Hand Custom", timeout=42 * 60),
            state_external=habapp_rules.actors.config.ventilation.StateConfig(level=103, display_text="External Custom"),
            state_humidity=habapp_rules.actors.config.ventilation.StateConfig(level=104, display_text="Humidity Custom"),
            state_long_absence=habapp_rules.actors.config.ventilation.StateConfigLongAbsence(level=105, display_text="Absence Custom", duration=1800, start_time=datetime.time(18)),
            after_run_timeout=350,
            current_threshold_power=0.5,
        )

        config_max = habapp_rules.actors.config.ventilation.VentilationTwoStageConfig(
            items=habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
                ventilation_output_on="Unittest_Ventilation_max_output_on",
                ventilation_output_power="Unittest_Ventilation_max_output_power",
                current="Unittest_Ventilation_max_current",
                manual="Unittest_Ventilation_max_manual",
                hand_request="Unittest_Ventilation_max_hand_request",
                external_request="Unittest_Ventilation_max_external_request",
                feedback_on="Unittest_Ventilation_max_feedback_on",
                feedback_power="Unittest_Ventilation_max_feedback_power",
                display_text="Unittest_Ventilation_max_display_text",
                presence_state="Unittest_Presence_state",
                state="Unittest_Ventilation_max_Custom_State",
            ),
            parameter=parameter_max,
        )

        config_min = habapp_rules.actors.config.ventilation.VentilationTwoStageConfig(
            items=habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
                ventilation_output_on="Unittest_Ventilation_min_output_on",
                ventilation_output_power="Unittest_Ventilation_min_output_power",
                current="Unittest_Ventilation_min_current",
                manual="Unittest_Ventilation_min_manual",
                state="H_Unittest_Ventilation_min_output_on_state",
            )
        )

        self.ventilation_min = habapp_rules.actors.ventilation.VentilationHeliosTwoStageHumidity(config_min)
        self.ventilation_max = habapp_rules.actors.ventilation.VentilationHeliosTwoStageHumidity(config_max)

    def test_init_without_current_item(self) -> None:
        """Test __init__ without current item."""
        config = habapp_rules.actors.config.ventilation.VentilationTwoStageConfig(
            items=habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
                ventilation_output_on="Unittest_Ventilation_min_output_on", ventilation_output_power="Unittest_Ventilation_min_output_power", manual="Unittest_Ventilation_min_manual", state="H_Unittest_Ventilation_min_output_on_state"
            )
        )
        with self.assertRaises(habapp_rules.core.exceptions.HabAppRulesConfigurationError):
            habapp_rules.actors.ventilation.VentilationHeliosTwoStageHumidity(config)

    def test_set_level(self) -> None:
        """Test _set_level."""
        TestCase = collections.namedtuple("TestCase", "state, expected_on, expected_power")

        test_cases = [
            TestCase("Manual", None, None),
            TestCase("Auto_PowerHand", "ON", "ON"),
            TestCase("Auto_Normal", "ON", "OFF"),
            TestCase("Auto_PowerExternal", "ON", "ON"),
            TestCase("Auto_LongAbsence_On", "ON", "ON"),
            TestCase("Auto_LongAbsence_Off", "OFF", "OFF"),
            TestCase("Auto_Init", None, None),
            TestCase("Auto_PowerAfterRun", "ON", "OFF"),
            TestCase("Auto_PowerHumidity", "ON", "OFF"),
        ]

        self.ventilation_max._config.parameter.state_normal.level = 1

        with unittest.mock.patch("habapp_rules.core.helper.send_if_different") as send_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    send_mock.reset_mock()
                    self.ventilation_max.state = test_case.state

                    self.ventilation_max._set_level()

                    if test_case.expected_on is not None:
                        send_mock.assert_any_call(self.ventilation_max._config.items.ventilation_output_on, test_case.expected_on)

                    if test_case.expected_power is not None:
                        send_mock.assert_any_call(self.ventilation_max._config.items.ventilation_output_power, test_case.expected_power)

    @unittest.skipIf(sys.platform != "win32", "Should only run on windows when graphviz is installed")
    def test_create_graph(self) -> None:  # pragma: no cover
        """Create state machine graph for documentation."""
        picture_dir = pathlib.Path(__file__).parent / "_state_charts" / "VentilationHeliosTwoStageHumidity"
        if not picture_dir.is_dir():
            picture_dir.mkdir(parents=True)

        graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(
            model=tests.helper.graph_machines.FakeModel(), states=self.ventilation_min.states, transitions=self.ventilation_min.trans, initial=self.ventilation_min.state, show_conditions=False
        )

        graph.get_graph().draw(picture_dir / "Ventilation.png", format="png", prog="dot")

        for state_name in [state for state in self._get_state_names(self.ventilation_min.states) if "init" not in state.lower()]:
            graph = tests.helper.graph_machines.HierarchicalGraphMachineTimer(model=tests.helper.graph_machines.FakeModel(), states=self.ventilation_min.states, transitions=self.ventilation_min.trans, initial=state_name, show_conditions=True)
            graph.get_graph(force_new=True, show_roi=True).draw(picture_dir / f"Ventilation_{state_name}.png", format="png", prog="dot")

    def test_get_initial_state(self) -> None:
        """Test _get_initial_state."""
        TestCase = collections.namedtuple("TestCase", "presence_state, current, manual, hand_request, external_request, expected_state_min, expected_state_max")

        test_cases = [
            # present | current = None
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, False, False, False, "Auto_Normal", "Auto_Normal"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, False, False, True, "Auto_Normal", "Auto_PowerExternal"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, False, True, False, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, False, True, True, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, True, False, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, True, False, True, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, True, True, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, None, True, True, True, "Manual", "Manual"),
            # present | current smaller than the threshold
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, False, False, False, "Auto_Normal", "Auto_Normal"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, False, False, True, "Auto_Normal", "Auto_PowerExternal"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, False, True, False, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, False, True, True, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, True, False, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, True, False, True, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, True, True, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 0.01, True, True, True, "Manual", "Manual"),
            # present | current greater than the threshold
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, False, False, False, "Auto_Normal", "Auto_PowerHumidity"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, False, False, True, "Auto_Normal", "Auto_PowerExternal"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, False, True, False, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, False, True, True, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, True, False, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, True, False, True, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, True, True, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.PRESENCE.value, 1, True, True, True, "Manual", "Manual"),
            # long absence
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, False, False, False, "Auto_Normal", "Auto_LongAbsence"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, False, False, True, "Auto_Normal", "Auto_LongAbsence"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, False, True, False, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, False, True, True, "Auto_Normal", "Auto_PowerHand"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, True, False, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, True, False, True, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, True, True, False, "Manual", "Manual"),
            TestCase(habapp_rules.system.PresenceState.LONG_ABSENCE.value, 20, True, True, True, "Manual", "Manual"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                tests.helper.oh_item.set_state("Unittest_Ventilation_min_manual", "ON" if test_case.manual else "OFF")
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_manual", "ON" if test_case.manual else "OFF")
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_current", test_case.current)
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_hand_request", "ON" if test_case.hand_request else "OFF")
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_external_request", "ON" if test_case.external_request else "OFF")
                tests.helper.oh_item.set_state("Unittest_Presence_state", test_case.presence_state)

                self.assertEqual(test_case.expected_state_min, self.ventilation_min._get_initial_state())
                self.assertEqual(test_case.expected_state_max, self.ventilation_max._get_initial_state())

    def test_set_feedback_states(self) -> None:
        """Test _set_feedback_states."""
        TestCase = collections.namedtuple("TestCase", "ventilation_level, state, expected_on, expected_power, expected_display_text")

        test_cases = [
            TestCase(None, "Auto_PowerHumidity", False, False, "Humidity Custom"),
            TestCase(0, "Auto_PowerHumidity", False, False, "Humidity Custom"),
            TestCase(1, "Auto_PowerHumidity", True, False, "Humidity Custom"),
            TestCase(2, "Auto_PowerHumidity", True, True, "Humidity Custom"),
        ]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.ventilation_min._ventilation_level = test_case.ventilation_level
                self.ventilation_max._ventilation_level = test_case.ventilation_level
                self.ventilation_min.state = test_case.state
                self.ventilation_max.state = test_case.state

                self.ventilation_min._set_feedback_states()
                self.ventilation_max._set_feedback_states()

                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_feedback_on", "ON" if test_case.expected_on else "OFF")
                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_feedback_power", "ON" if test_case.expected_power else "OFF")
                tests.helper.oh_item.assert_value("Unittest_Ventilation_max_display_text", test_case.expected_display_text)

    def test_current_greater_threshold(self) -> None:
        """Test __current_greater_threshold."""
        TestCase = collections.namedtuple("TestCase", "threshold, item_value, given_value, expected_result")

        test_cases = [TestCase(42, 0, 0, False), TestCase(42, 0, 100, True), TestCase(42, 100, 0, False), TestCase(42, 100, 100, True), TestCase(42, 0, None, False), TestCase(42, 100, None, True), TestCase(42, None, None, False)]

        for test_case in test_cases:
            with self.subTest(test_case=test_case):
                self.ventilation_max._current_threshold_power = test_case.threshold
                tests.helper.oh_item.set_state("Unittest_Ventilation_max_current", test_case.item_value)

                result = self.ventilation_max._current_greater_threshold() if test_case.given_value is None else self.ventilation_max._current_greater_threshold(test_case.given_value)

                self.assertEqual(test_case.expected_result, result)

    def test_power_after_run_transitions(self) -> None:
        """Test transitions of PowerAfterRun."""
        # _end_after_run triggered
        self.ventilation_max.to_Auto_PowerAfterRun()
        self.ventilation_max._end_after_run()
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

    def test_power_humidity_transitions(self) -> None:
        """Test transitions of state Auto_PowerHumidity."""
        # set default config parameters
        self.ventilation_min._config.parameter = habapp_rules.actors.config.ventilation.VentilationTwoStageParameter()
        self.ventilation_max._config.parameter = habapp_rules.actors.config.ventilation.VentilationTwoStageParameter()

        # set AutoNormal as initial state
        self.ventilation_min.to_Auto_Normal()
        self.ventilation_max.to_Auto_Normal()

        # set correct output states
        self.ventilation_min._config.items.ventilation_output_power.set_value("OFF")
        self.ventilation_max._config.items.ventilation_output_power.set_value("OFF")

        # state != Auto_PowerHumidity | current below the threshold
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_min_current", 0.1)
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_max_current", 0.1)

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

        tests.helper.oh_item.assert_value("Unittest_Ventilation_min_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_min_output_power", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_power", "OFF")

        # state != Auto_PowerHumidity | current grater then the threshold
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_min_current", 0.2)
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_max_current", 0.6)

        self.assertEqual("Auto_PowerHumidity", self.ventilation_min.state)
        self.assertEqual("Auto_PowerHumidity", self.ventilation_max.state)

        tests.helper.oh_item.assert_value("Unittest_Ventilation_min_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_min_output_power", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_power", "OFF")

        # state == Auto_PowerHumidity | current grater then the threshold
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_min_current", 0.2)
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_max_current", 0.6)

        self.assertEqual("Auto_PowerHumidity", self.ventilation_min.state)
        self.assertEqual("Auto_PowerHumidity", self.ventilation_max.state)

        # state == Auto_PowerHumidity | current below then the threshold
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_min_current", 0.1)
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_max_current", 0.1)

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

        # state == Auto_PowerAfterRun | current below the threshold
        self.ventilation_min.to_Auto_PowerAfterRun()
        self.ventilation_max.to_Auto_PowerAfterRun()

        tests.helper.oh_item.item_state_event("Unittest_Ventilation_min_current", 0.1)
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_max_current", 0.1)

        self.ventilation_min._after_run_timeout()
        self.ventilation_max._after_run_timeout()

        self.assertEqual("Auto_Normal", self.ventilation_min.state)
        self.assertEqual("Auto_Normal", self.ventilation_max.state)

        # state == Auto_PowerAfterRun | current grater then the threshold
        self.ventilation_min.to_Auto_PowerAfterRun()
        self.ventilation_max.to_Auto_PowerAfterRun()

        tests.helper.oh_item.item_state_event("Unittest_Ventilation_min_current", 0.2)
        tests.helper.oh_item.item_state_event("Unittest_Ventilation_max_current", 0.6)

        self.ventilation_min._after_run_timeout()
        self.ventilation_max._after_run_timeout()

        self.assertEqual("Auto_PowerHumidity", self.ventilation_min.state)
        self.assertEqual("Auto_PowerHumidity", self.ventilation_max.state)

        tests.helper.oh_item.assert_value("Unittest_Ventilation_min_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_min_output_power", "OFF")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_power", "OFF")

        # state == Auto_PowerHumidity | _hand_on triggered
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "ON")
        self.assertEqual("Auto_PowerHand", self.ventilation_max.state)

        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_power", "ON")

        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_hand_request", "OFF")

        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_on", "ON")
        tests.helper.oh_item.assert_value("Unittest_Ventilation_max_output_power", "OFF")

        # state == Auto_PowerHumidity | _external_on triggered
        self.ventilation_max.to_Auto_PowerHumidity()
        tests.helper.oh_item.item_state_change_event("Unittest_Ventilation_max_external_request", "ON")
        self.assertEqual("Auto_PowerExternal", self.ventilation_max.state)



"""Watchdog rules."""

import HABApp

import habapp_rules.core.helper
from habapp_rules.system.config.item_watchdog import WatchdogConfig


class ItemWatchdog(HABApp.Rule):
    """Watchdog rule to check if the observed item was updated in time.

    # Items:
    Switch    Item_To_Observe   "Item which should be observed"
    Switch    Warning           "Warning, item was not updated in time"

    # Config:
    config = habapp_rules.system.config.WatchdogConfig(
            items=habapp_rules.system.config.WatchdogItems(
                    observed="Item_To_Observe",
                    warning="Warning")
    )

    # Rule init:
    habapp_rules.system.watchdog.Watchdog(config)
    """

    def __init__(self, config: WatchdogConfig) -> None:
        """Init watchdog rule.

        Args:
            config: Config for watchdog rule
        """
        HABApp.Rule.__init__(self)
        self._config = config

        self._countdown = self.run.countdown(self._config.parameter.timeout, habapp_rules.core.helper.send_if_different, item=self._config.items.warning, value="ON")
        self._countdown.reset()
        self._config.items.observed.listen_event(self._cb_observed_state_updated, HABApp.openhab.events.ItemStateUpdatedEventFilter())

    def _cb_observed_state_updated(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:  # noqa: ARG002
        """Callback which is called if the observed item was updated.

        Args:
            event: event which triggered this callback
        """
        habapp_rules.core.helper.send_if_different(self._config.items.warning, "OFF")
        self._countdown.reset()



"""Ventilation rules."""

import abc
import copy
import datetime
import logging
import typing

import HABApp

import habapp_rules.actors.config.ventilation
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system
from habapp_rules.actors.config.ventilation import VentilationTwoStageItems
from habapp_rules.core.helper import send_if_different

LOGGER = logging.getLogger(__name__)

LEVEL_POWER = 2


def _to_datetime(time_input: datetime.time) -> datetime.datetime:  # this is needed because of https://github.com/spacemanspiff2007/eascheduler/issues
    """Converts a datetime.time object to a datetime.datetime object using today's date. Adds one day if the time is in the past.

    Args:
        time_input: The time to convert.

    Returns:
        The resulting datetime object with today's date or the next day if the time is in the past.
    """
    result = datetime.datetime.combine(datetime.datetime.now(), time_input)

    # If the resulting datetime is in the past, add one day
    if result <= datetime.datetime.now():
        result += datetime.timedelta(days=1)

    return result


class _VentilationBase(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Class for ventilation objects."""

    _config: VentilationTwoStageItems

    states: typing.ClassVar = [
        {"name": "Manual"},
        {
            "name": "Auto",
            "initial": "Init",
            "children": [
                {"name": "Init"},
                {"name": "Normal"},
                {"name": "PowerHand", "timeout": 3600, "on_timeout": "_hand_off"},
                {"name": "PowerExternal"},
                {"name": "LongAbsence", "initial": "Off", "children": [{"name": "On", "timeout": 3600, "on_timeout": "_long_absence_power_off"}, {"name": "Off"}]},
            ],
        },
    ]

    trans: typing.ClassVar = [
        # manual
        {"trigger": "_manual_on", "source": ["Auto"], "dest": "Manual"},
        {"trigger": "_manual_off", "source": "Manual", "dest": "Auto"},
        # PowerHand
        {"trigger": "_hand_on", "source": ["Auto_Normal", "Auto_PowerExternal", "Auto_LongAbsence"], "dest": "Auto_PowerHand"},
        {"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_PowerExternal", "conditions": "_external_active_and_configured"},
        {"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_Normal", "unless": "_external_active_and_configured"},
        # PowerExternal
        {"trigger": "_external_on", "source": "Auto_Normal", "dest": "Auto_PowerExternal"},
        {"trigger": "_external_off", "source": "Auto_PowerExternal", "dest": "Auto_Normal"},
        # long absence
        {"trigger": "_long_absence_on", "source": ["Auto_Normal", "Auto_PowerExternal"], "dest": "Auto_LongAbsence"},
        {"trigger": "_long_absence_power_on", "source": "Auto_LongAbsence_Off", "dest": "Auto_LongAbsence_On"},
        {"trigger": "_long_absence_power_off", "source": "Auto_LongAbsence_On", "dest": "Auto_LongAbsence_Off"},
        {"trigger": "_long_absence_off", "source": "Auto_LongAbsence", "dest": "Auto_Normal"},
    ]

    def __init__(self, config: habapp_rules.actors.config.ventilation.VentilationConfig) -> None:
        """Init of ventilation base.

        Args:
            config: ventilation config
        """
        self._config = config
        self._ventilation_level: int | None = None

        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)

        # init state machine
        self._previous_state = None
        self._state_change_time = datetime.datetime.now()
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_initial_state()

        self._apply_config()

        # callbacks
        self._config.items.manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.hand_request is not None:
            self._config.items.hand_request.listen_event(self._cb_power_hand_request, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.external_request is not None:
            self._config.items.external_request.listen_event(self._cb_external_request, HABApp.openhab.events.ItemStateChangedEventFilter())
        if self._config.items.presence_state is not None:
            self._config.items.presence_state.listen_event(self._cb_presence_state, HABApp.openhab.events.ItemStateChangedEventFilter())

        self._update_openhab_state()

    def _get_initial_state(self, default_value: str = "initial") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        if self._config.items.manual.is_on():
            return "Manual"
        if self._config.items.hand_request is not None and self._config.items.hand_request.is_on():
            return "Auto_PowerHand"
        if self._config.items.presence_state is not None and self._config.items.presence_state.value == habapp_rules.system.PresenceState.LONG_ABSENCE.value:
            return "Auto_LongAbsence"
        if self._config.items.external_request is not None and self._config.items.external_request.is_on():
            return "Auto_PowerExternal"
        return "Auto_Normal"

    def _apply_config(self) -> None:
        """Apply values from config."""
        self.state_machine.get_state("Auto_PowerHand").timeout = self._config.parameter.state_hand.timeout
        self.state_machine.get_state("Auto_LongAbsence_On").timeout = self._config.parameter.state_long_absence.duration

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states.

        This method should be set to "after_state_change" of the state machine.
        """
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self._state_change_time = datetime.datetime.now()
            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

            self._set_level()
            self._set_feedback_states()
            self._previous_state = self.state

    def _set_level(self) -> None:
        """Set ventilation level."""
        if self.state == "Manual":
            return

        if self.state == "Auto_PowerHand":
            self._ventilation_level = self._config.parameter.state_hand.level
        elif self.state == "Auto_Normal":
            self._ventilation_level = self._config.parameter.state_normal.level
        elif self.state == "Auto_PowerExternal":
            self._ventilation_level = self._config.parameter.state_external.level
        elif self.state == "Auto_LongAbsence_On":
            self._ventilation_level = self._config.parameter.state_long_absence.level
        elif self.state == "Auto_LongAbsence_Off":
            self._ventilation_level = 0
        else:
            return

        self._set_level_to_ventilation_items()

    @abc.abstractmethod
    def _set_level_to_ventilation_items(self) -> None:
        """Set ventilation to output item(s)."""

    def _get_display_text(self) -> str | None:
        """Get Text for display.

        Returns:
             text for display or None if not defined for this state
        """
        if self.state == "Manual":
            return "Manual"
        if self.state == "Auto_Normal":
            return self._config.parameter.state_normal.display_text
        if self.state == "Auto_PowerExternal":
            return self._config.parameter.state_external.display_text
        if self.state == "Auto_LongAbsence_On":
            return f"{self._config.parameter.state_long_absence.display_text} ON"
        if self.state == "Auto_LongAbsence_Off":
            return f"{self._config.parameter.state_long_absence.display_text} OFF"

        return None

    def _set_feedback_states(self) -> None:
        """Set feedback sates to the OpenHAB items."""
        if self._config.items.hand_request is not None and self._previous_state == "Auto_PowerHand":
            habapp_rules.core.helper.send_if_different(self._config.items.hand_request, "OFF")

        if self._config.items.feedback_on is not None:
            habapp_rules.core.helper.send_if_different(self._config.items.feedback_on, "ON" if self._ventilation_level else "OFF")

        if self._config.items.feedback_power is not None:
            target_value = "ON" if self._ventilation_level is not None and self._ventilation_level >= LEVEL_POWER else "OFF"
            habapp_rules.core.helper.send_if_different(self._config.items.feedback_power, target_value)

        if self._config.items.display_text is not None:
            if self.state == "Auto_PowerHand":
                self.__set_hand_display_text()
                return

            if (display_text := self._get_display_text()) is not None:
                habapp_rules.core.helper.send_if_different(self._config.items.display_text, display_text)

    def __set_hand_display_text(self) -> None:
        """Callback to set display text."""
        if self.state != "Auto_PowerHand":
            # state changed and is not PowerHand anymore
            return

        # get the remaining minutes and set display text
        remaining_minutes = round((self._config.parameter.state_hand.timeout - (datetime.datetime.now() - self._state_change_time).seconds) / 60)
        remaining_minutes = max(remaining_minutes, 0)
        habapp_rules.core.helper.send_if_different(self._config.items.display_text, f"{self._config.parameter.state_hand.display_text} {remaining_minutes}min")

        # re-trigger this method in 1 minute
        self.run.once(60, self.__set_hand_display_text)

    def on_enter_Auto_Init(self) -> None:  # noqa: N802
        """Is called on entering of Auto_Init state."""
        self._set_initial_state()

    def on_enter_Auto_LongAbsence_Off(self) -> None:  # noqa: N802
        """Is called on entering of Auto_LongAbsence_Off state."""
        trigger_time = _to_datetime(self._config.parameter.state_long_absence.start_time)
        self.run.once(trigger_time, self._trigger_long_absence_power_on)

    def _trigger_long_absence_power_on(self) -> None:
        """Trigger long absence power on."""
        self._long_absence_power_on()

    def _external_active_and_configured(self) -> bool:
        """Check if external request is active and configured.

        Returns:
            True if external request is active and configured
        """
        return self._config.items.external_request is not None and self._config.items.external_request.is_on()

    def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if manual mode changed.

        Args:
            event: original trigger event
        """
        if event.value == "ON":
            self._manual_on()
        else:
            self._manual_off()

    def _cb_power_hand_request(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if power_hand_request changed.

        Args:
            event: original trigger event
        """
        if event.value == "ON":
            self._hand_on()
        else:
            self._hand_off()

    def _cb_external_request(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if external request changed.

        Args:
            event: original trigger event
        """
        if event.value == "ON":
            self._external_on()
        else:
            self._external_off()

    def _cb_presence_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if presence_state changed.

        Args:
            event: original trigger event
        """
        if event.value == habapp_rules.system.PresenceState.LONG_ABSENCE.value:
            self._long_absence_on()
        else:
            self._long_absence_off()


class Ventilation(_VentilationBase):
    """Rule for managing ventilation systems which can be controlled with ventilation levels.

    # Items:
    Number  Ventilation_level           "Ventilation level"
    Switch  Manual                      "Manual"
    Switch  Hand_Request                "Hand request"
    Switch  External_Request            "External request"
    String  presence_state              "Presence state"
    Switch  Feedback_On                 "Feedback is ON"
    Switch  Feedback_Power              "Feedback is Power"

    # Config
    config = habapp_rules.actors.config.ventilation.VentilationConfig(
            items=habapp_rules.actors.config.ventilation.VentilationItems(
                    ventilation_level="Ventilation_level",
                    manual="Manual",
                    hand_request="Hand_Request",
                    external_request="External_Request",
                    presence_state="presence_state",
                    feedback_on="Feedback_On",
                    feedback_power="Feedback_Power"
            )
    )

    # Rule init:
    habapp_rules.actors.ventilation.Ventilation(config)
    """

    def __init__(self, config: habapp_rules.actors.config.ventilation.VentilationConfig) -> None:
        """Init of ventilation object.

        Args:
            config: config of the ventilation rule
        """
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.ventilation_level.name)

        _VentilationBase.__init__(self, config)
        self._instance_logger.info(habapp_rules.core.state_machine_rule.StateMachineRule.get_initial_log_message(self))

    def _set_level_to_ventilation_items(self) -> None:
        """Set ventilation to output item(s)."""
        habapp_rules.core.helper.send_if_different(self._config.items.ventilation_level, self._ventilation_level)


class VentilationHeliosTwoStage(_VentilationBase):
    """Rule for managing Helios ventilation systems with humidity sensor (E.g. Helios ELS).

    # Items:
    Switch  Ventilation_Switch_On       "Ventilation relay on"
    Switch  Ventilation_Switch_Power    "Ventilation relay power"
    Switch  Manual                      "Manual"
    Switch  Hand_Request                "Hand request"
    Switch  External_Request            "External request"
    String  presence_state              "Presence state"
    Switch  Feedback_On                 "Feedback is ON"
    Switch  Feedback_Power              "Feedback is Power"

    # Config
    config = habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
            items=habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
                    ventilation_output_on="Ventilation_Switch_On",
                    ventilation_output_power="Ventilation_Switch_Power",
                    manual="Manual",
                    hand_request="Hand_Request",
                    external_request="External_Request",
                    presence_state="presence_state",
                    feedback_on="Feedback_On",
                    feedback_power="Feedback_Power"
            )
    )

    # Rule init:
    habapp_rules.actors.ventilation.VentilationHeliosTwoStage(config)
    """

    states = copy.deepcopy(_VentilationBase.states)
    __AUTO_STATE = next(state for state in states if state["name"] == "Auto")  # pragma: no cover
    __AUTO_STATE["children"].append({"name": "PowerAfterRun", "timeout": 390, "on_timeout": "_after_run_timeout"})

    trans = copy.deepcopy(_VentilationBase.trans)
    # remove not needed transitions
    trans.remove({"trigger": "_hand_on", "source": ["Auto_Normal", "Auto_PowerExternal", "Auto_LongAbsence"], "dest": "Auto_PowerHand"})  # will be extended with additional source state
    trans.remove({"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_Normal", "unless": "_external_active_and_configured"})  # this is not needed anymore since there is always Auto_PowerAfterRun after any power state
    trans.remove({"trigger": "_external_on", "source": "Auto_Normal", "dest": "Auto_PowerExternal"})  # will be extended with additional source state
    trans.remove({"trigger": "_external_off", "source": "Auto_PowerExternal", "dest": "Auto_Normal"})  # this is not needed anymore since there is always Auto_PowerAfterRun after any power state

    # add new PowerHand transitions
    trans.append({"trigger": "_hand_on", "source": ["Auto_Normal", "Auto_PowerExternal", "Auto_LongAbsence", "Auto_PowerAfterRun"], "dest": "Auto_PowerHand"})
    trans.append({"trigger": "_hand_off", "source": "Auto_PowerHand", "dest": "Auto_PowerAfterRun", "unless": "_external_active_and_configured"})

    # add new PowerExternal transitions
    trans.append({"trigger": "_external_on", "source": ["Auto_Normal", "Auto_PowerAfterRun"], "dest": "Auto_PowerExternal"})
    trans.append({"trigger": "_external_off", "source": "Auto_PowerExternal", "dest": "Auto_PowerAfterRun"})

    # add new PowerAfterRun transitions
    trans.append({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal"})

    def __init__(self, config: habapp_rules.actors.config.ventilation.VentilationTwoStageConfig) -> None:
        """Init of a Helios ventilation object which uses two switches to set the level.

        Args:
            config: config for the ventilation rule
        """
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, config.items.ventilation_output_on.name)
        _VentilationBase.__init__(self, config)

        # set timeout
        self.state_machine.get_state("Auto_PowerAfterRun").timeout = config.parameter.after_run_timeout

        self._instance_logger.info(habapp_rules.core.state_machine_rule.StateMachineRule.get_initial_log_message(self))

    def _get_display_text(self) -> str | None:
        """Get Text for display.

        Returns:
            text for display or None if not defined for this state
        """
        if self.state == "Auto_PowerAfterRun":
            return self._config.parameter.state_after_run.display_text

        return _VentilationBase._get_display_text(self)  # noqa: SLF001

    def _set_level(self) -> None:
        """Set ventilation level."""
        if self.state == "Auto_PowerAfterRun":
            self._ventilation_level = self._config.parameter.state_after_run.level
            self._set_level_to_ventilation_items()

        else:
            super()._set_level()

        if self._config.items.feedback_ventilation_level is not None:
            send_if_different(self._config.items.feedback_ventilation_level, self._ventilation_level)

    def _set_level_to_ventilation_items(self) -> None:
        """Set ventilation to output item(s)."""
        if self.state == "Auto_PowerAfterRun":
            habapp_rules.core.helper.send_if_different(self._config.items.ventilation_output_on, "ON")
            habapp_rules.core.helper.send_if_different(self._config.items.ventilation_output_power, "OFF")

        else:
            habapp_rules.core.helper.send_if_different(self._config.items.ventilation_output_on, "ON" if self._ventilation_level else "OFF")
            habapp_rules.core.helper.send_if_different(self._config.items.ventilation_output_power, "ON" if self._ventilation_level >= LEVEL_POWER else "OFF")


class VentilationHeliosTwoStageHumidity(VentilationHeliosTwoStage):
    """Rule for managing Helios ventilation systems with humidity sensor (E.g. Helios ELS).

    # Items:
    Switch  Ventilation_Switch_On       "Ventilation relay on"
    Switch  Ventilation_Switch_Power    "Ventilation relay power"
    Number  Ventilation_Current         "Ventilation current"
    Switch  Manual                      "Manual"
    Switch  Hand_Request                "Hand request"
    Switch  External_Request            "External request"
    String  presence_state              "Presence state"
    Switch  Feedback_On                 "Feedback is ON"
    Switch  Feedback_Power              "Feedback is Power"

    # Config
    config = habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
            items=habapp_rules.actors.config.ventilation.VentilationTwoStageItems(
                    ventilation_output_on="Ventilation_Switch_On",
                    ventilation_output_power="Ventilation_Switch_Power",
                    current="Ventilation_Current",
                    manual="Manual",
                    hand_request="Hand_Request",
                    external_request="External_Request",
                    presence_state="presence_state",
                    feedback_on="Feedback_On",
                    feedback_power="Feedback_Power"
            )
    )

    # Rule init:
    habapp_rules.actors.ventilation.VentilationHeliosTwoStageHumidity(config)
    """

    states = copy.deepcopy(VentilationHeliosTwoStage.states)
    __AUTO_STATE = next(state for state in states if state["name"] == "Auto")  # pragma: no cover
    __AUTO_STATE["children"].append({"name": "PowerHumidity"})

    trans = copy.deepcopy(VentilationHeliosTwoStage.trans)
    # remove not needed transitions
    trans.remove({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal"})  # will be changed to only go to AutoNormal if the current is below the threshold (not humidity)

    # add new PowerHumidity transitions
    trans.append({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal", "unless": "_current_greater_threshold"})
    trans.append({"trigger": "_end_after_run", "source": "Auto_PowerAfterRun", "dest": "Auto_Normal"})
    trans.append({"trigger": "_after_run_timeout", "source": "Auto_PowerAfterRun", "dest": "Auto_PowerHumidity", "conditions": "_current_greater_threshold"})

    trans.append({"trigger": "_humidity_on", "source": "Auto_Normal", "dest": "Auto_PowerHumidity"})
    trans.append({"trigger": "_humidity_off", "source": "Auto_PowerHumidity", "dest": "Auto_Normal"})

    trans.append({"trigger": "_hand_on", "source": "Auto_PowerHumidity", "dest": "Auto_PowerHand"})
    trans.append({"trigger": "_external_on", "source": "Auto_PowerHumidity", "dest": "Auto_PowerExternal"})

    def __init__(self, config: habapp_rules.actors.config.ventilation.VentilationTwoStageConfig) -> None:
        """Init of a Helios ventilation object which uses two switches to set the level, including a humidity sensor.

        Args:
            config: configuration of the ventilation rule

        Raises:
            habapp_rules.core.exceptions.HabAppRulesConfigurationError: if config is missing required items
        """
        if config.items.current is None:
            msg = "Missing item 'current'"
            raise habapp_rules.core.exceptions.HabAppRulesConfigurationError(msg)
        self._current_threshold_power = config.parameter.current_threshold_power

        VentilationHeliosTwoStage.__init__(self, config)
        config.items.current.listen_event(self._cb_current, HABApp.openhab.events.ItemStateUpdatedEventFilter())

    def _get_initial_state(self, default_value: str = "initial") -> str:
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        state = super()._get_initial_state(default_value)

        if state == "Auto_Normal" and self._current_greater_threshold():
            return "Auto_PowerHumidity"

        return state

    def _get_display_text(self) -> str | None:
        """Get Text for display.

        Returns:
            text for display or None if not defined for this state
        """
        if self.state == "Auto_PowerHumidity":
            return self._config.parameter.state_humidity.display_text

        return VentilationHeliosTwoStage._get_display_text(self)  # noqa: SLF001

    def _set_level(self) -> None:
        """Set ventilation level."""
        if self.state == "Auto_PowerHumidity":
            self._ventilation_level = self._config.parameter.state_humidity.level
            self._set_level_to_ventilation_items()
            return

        super()._set_level()

    def _set_level_to_ventilation_items(self) -> None:
        """Set ventilation to output item(s)."""
        if self.state == "Auto_PowerHumidity":
            habapp_rules.core.helper.send_if_different(self._config.items.ventilation_output_on, "ON")
            habapp_rules.core.helper.send_if_different(self._config.items.ventilation_output_power, "OFF")
        else:
            super()._set_level_to_ventilation_items()

    def _current_greater_threshold(self, current: float | None = None) -> bool:
        """Check if current is greater than the threshold.

        Args:
            current: current which should be checked. If None the value of the current item will be taken

        Returns:
            True if current greater than the threshold, else False
        """
        current = current if current is not None else self._config.items.current.value

        if current is None:
            return False

        return current > self._current_threshold_power

    def _cb_current(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:
        """Callback which is triggered if the current changed.

        Args:
            event: original trigger event
        """
        if self.state != "Auto_PowerHumidity" and self._current_greater_threshold(event.value):
            self._humidity_on()
        elif self.state == "Auto_PowerHumidity" and not self._current_greater_threshold(event.value):
            self._humidity_off()
        elif self.state == "Auto_PowerAfterRun" and not self._current_greater_threshold(event.value):
            self._end_after_run()



"""Rules for astro actions."""

import abc
import logging

import HABApp

import habapp_rules.core.helper
import habapp_rules.sensors.config.astro

LOGGER = logging.getLogger(__name__)


class _SetNightDayBase(HABApp.Rule):
    """Base class for set night / day."""

    def __init__(self, item_target: HABApp.openhab.items.SwitchItem, item_elevation: HABApp.openhab.items.NumberItem, elevation_threshold: float) -> None:
        """Init Rule.

        Args:
            item_target: OpenHab item which should be set depending on the sun elevation value
            item_elevation: OpenHAB item of sun elevation (NumberItem)
            elevation_threshold: Threshold value for elevation.
        """
        HABApp.Rule.__init__(self)

        self._item_target = item_target
        self._item_elevation = item_elevation
        self._elevation_threshold = elevation_threshold

        self._item_elevation.listen_event(self._set_night, HABApp.openhab.events.ItemStateChangedEventFilter())

        self.run.soon(self._set_night)

    def _set_night(self, _: HABApp.openhab.events.ItemStateChangedEvent | None = None) -> None:
        """Callback which sets the state to the night item."""
        if self._item_elevation.value is None:
            return
        habapp_rules.core.helper.send_if_different(self._item_target, self._get_target_value())

    @abc.abstractmethod
    def _get_target_value(self) -> str:
        """Get target value which should be set.

        Returns:
            target value (ON / OFF)
        """


class SetDay(_SetNightDayBase):
    """Rule to set / unset day item at dusk / dawn.

    # Items:
    Switch    day                   "Day"
    Number    elevation             "Sun elevation"    <sun>     {channel="astro...}

    # Config:
    config = habapp_rules.sensors.config.astro.SetDayConfig(
            items=habapp_rules.sensors.config.astro.SetDayItems(
                    day="day",
                    elevation="elevation"
            ),
            parameter=habapp_rules.sensors.config.astro.SetDayParameter(
                    elevation_threshold=5
            )
    )

    # Rule init:
    habapp_rules.sensors.astro.SetNight(config)
    """

    def __init__(self, config: habapp_rules.sensors.config.astro.SetDayConfig) -> None:
        """Init Rule.

        Args:
            config: Config for set day rule
        """
        _SetNightDayBase.__init__(self, config.items.day, config.items.elevation, config.parameter.elevation_threshold)

    def _get_target_value(self) -> str:
        """Get target value which should be set.

        Returns:
            target value (ON / OFF)
        """
        return "ON" if self._item_elevation.value > self._elevation_threshold else "OFF"


class SetNight(_SetNightDayBase):
    """Rule to set / unset night item at dusk / dawn.

    # Items:
    Switch    night_for_shading     "Night for shading"
    Number    elevation             "Sun elevation"    <sun>     {channel="astro...}

    # Config:
    config = habapp_rules.sensors.config.astro.SetNightConfig(
            items=habapp_rules.sensors.config.astro.SetNightItems(
                    night="night_for_shading",
                    elevation="elevation"
            ),
            parameter=habapp_rules.sensors.config.astro.SetNightParameter(
                    elevation_threshold=5
            )
    )

    # Rule init:
    habapp_rules.sensors.astro.SetNight(config)
    """

    def __init__(self, config: habapp_rules.sensors.config.astro.SetNightConfig) -> None:
        """Init Rule.

        Args:
            config: Config for setting night depending on sun elevation
        """
        _SetNightDayBase.__init__(self, config.items.night, config.items.elevation, config.parameter.elevation_threshold)

    def _get_target_value(self) -> str:
        """Get target value which should be set.

        Returns:
            target value (ON / OFF)
        """
        return "ON" if self._item_elevation.value < self._elevation_threshold else "OFF"



"""Unit-test for state_machine."""

import collections
import time
import unittest
import unittest.mock

import HABApp.openhab.items.switch_item

import habapp_rules.core.exceptions
import habapp_rules.core.state_machine_rule
import tests.helper.oh_item
import tests.helper.test_case_base


class TestStateMachineRule(tests.helper.test_case_base.TestCaseBase):
    """Tests for StateMachineRule."""

    def setUp(self) -> None:
        """Setup unit-tests."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.StringItem, "Unittest_State", None)
        self.state_item = HABApp.openhab.items.StringItem.get_item("Unittest_State")

        self.item_exists_mock.return_value = False

        with unittest.mock.patch("habapp_rules.core.helper.create_additional_item", return_value=HABApp.openhab.items.string_item.StringItem("rules_common_state_machine_rule_StateMachineRule_state", "")):
            self._state_machine = habapp_rules.core.state_machine_rule.StateMachineRule(self.state_item)

    def test_get_initial_state(self) -> None:
        """Test getting of initial state."""
        TestCase = collections.namedtuple("TestCase", "item_value, state_names, default, expected_result")
        test_cases = [
            TestCase("state1", ["state1", "state2"], "default", "state1"),
            TestCase("wrong_state", ["state1", "state2"], "default", "default"),
            TestCase("state1", ["new_state1", "new_state_2"], "default", "default"),
            TestCase("state1", [], "default", "default"),
        ]

        with unittest.mock.patch.object(self._state_machine, "_item_state") as state_item_mock:
            for test_case in test_cases:
                state_item_mock.value = test_case.item_value
                self._state_machine.states = [{"name": name} for name in test_case.state_names]
                self.assertEqual(self._state_machine._get_initial_state(test_case.default), test_case.expected_result)

    def test_update_openhab_state(self) -> None:
        """Test if OpenHAB state will be updated."""
        self._state_machine.state = "some_state"
        with unittest.mock.patch.object(self._state_machine, "_item_state") as state_item:
            self._state_machine._update_openhab_state()
            state_item.oh_send_command.assert_called_once_with("some_state")

    def test_on_rule_removed(self) -> None:
        """Test on_rule_removed."""
        # check if 'on_rule_removed' is still available in HABApp
        self.assertIsNotNone(HABApp.rule.Rule.on_rule_removed)

        # check if timer is stopped correctly
        states = [{"name": "stopped"}, {"name": "running", "timeout": 99, "on_timeout": "trigger_stop"}]

        with unittest.mock.patch("habapp_rules.core.helper.create_additional_item", return_value=HABApp.openhab.items.string_item.StringItem("rules_common_state_machine_rule_StateMachineRule_state", "")):
            for initial_state in ["stopped", "running"]:
                state_machine_rule = habapp_rules.core.state_machine_rule.StateMachineRule(self.state_item)

                state_machine_rule.state_machine = habapp_rules.core.state_machine_rule.StateMachineWithTimeout(model=state_machine_rule, states=states, ignore_invalid_triggers=True)

                state_machine_rule._set_state(initial_state)

                if initial_state == "running":
                    self.assertTrue(next(iter(state_machine_rule.state_machine.states["running"].runner.values())).is_alive())

                state_machine_rule.on_rule_removed()

                if initial_state == "running":
                    time.sleep(0.001)
                    self.assertFalse(next(iter(state_machine_rule.state_machine.states["running"].runner.values())).is_alive())



"""Set version rules."""

import logging
import platform

import HABApp

import habapp_rules
import habapp_rules.core.helper

LOGGER = logging.getLogger(__name__)


class SetVersions(HABApp.Rule):
    """Update HABApp and habapp_rules version to OpenHAB items."""

    def __init__(self) -> None:
        """Init rule."""
        HABApp.Rule.__init__(self)
        LOGGER.info("Update versions of OpenHAB items")

        item_version_habapp = habapp_rules.core.helper.create_additional_item("H_habapp_version", "String", "HABApp version")
        item_version_habapp_rules = habapp_rules.core.helper.create_additional_item("H_habapp_rules_version", "String", "habapp_rules version")
        item_version_python = habapp_rules.core.helper.create_additional_item("H_python_version", "String", "Python version")

        item_version_habapp.oh_send_command(HABApp.__version__)
        item_version_habapp_rules.oh_send_command(habapp_rules.__version__)
        item_version_python.oh_send_command(platform.python_version())



"""Bathroom light rules."""

import logging
import time
import typing

import HABApp

import habapp_rules.actors.config.energy_save_switch
import habapp_rules.actors.config.light_bathroom
import habapp_rules.actors.state_observer
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system

LOGGER = logging.getLogger(__name__)


class BathroomLight(habapp_rules.core.state_machine_rule.StateMachineRule):
    """Bathroom light rule."""

    states: typing.ClassVar = [
        {"name": "Manual"},
        {
            "name": "Auto",
            "initial": "Init",
            "children": [
                {"name": "Init"},
                {"name": "Off"},
                {"name": "On", "initial": "Init", "children": [{"name": "Init"}, {"name": "MainDay"}, {"name": "MainNight"}, {"name": "MainAndMirror"}]},
            ],
        },
    ]

    trans: typing.ClassVar = [
        # manual
        {"trigger": "manual_on", "source": "Auto", "dest": "Manual"},
        {"trigger": "manual_off", "source": "Manual", "dest": "Auto"},
        # switch on
        {"trigger": "hand_on", "source": "Auto_Off", "dest": "Auto_On"},
        # mirror
        {"trigger": "mirror_on", "source": ["Auto_On_MainDay", "Auto_On_MainNight"], "dest": "Auto_On_MainAndMirror"},
        {"trigger": "mirror_off", "source": "Auto_On_MainAndMirror", "dest": "Auto_On_MainDay", "conditions": "_is_day"},
        {"trigger": "mirror_off", "source": "Auto_On_MainAndMirror", "dest": "Auto_On_MainNight", "unless": "_is_day"},
        # off
        {"trigger": "hand_off", "source": "Auto_On", "dest": "Auto_Off"},
        {"trigger": "sleep_started", "source": "Auto_On", "dest": "Auto_Off"},
        {"trigger": "leaving", "source": "Auto_On", "dest": "Auto_Off"},
    ]

    def __init__(self, config: habapp_rules.actors.config.light_bathroom.BathroomLightConfig) -> None:
        """Init rule.

        Args:
            config: Config of bathroom light rule
        """
        self._config = config
        habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, self._config.items.state)
        self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self._config.items.light_main.name)

        self._sleep_end_time = 0
        self._switch_on_via_increase = False
        control_names = [self._config.items.light_main_ctr.name] if self._config.items.light_main_ctr is not None else []
        self._light_main_observer = habapp_rules.actors.state_observer.StateObserverDimmer(self._config.items.light_main.name, control_names=control_names, cb_on=self._cb_hand_on, cb_off=self._cb_hand_off, value_tolerance=5)

        # init state machine
        self._previous_state = None
        self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(model=self, states=self.states, transitions=self.trans, ignore_invalid_triggers=True, after_state_change="_update_openhab_state")
        self._set_state(self._get_initial_state())

        # callbacks
        self._config.items.manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
        self._config.items.light_mirror.listen_event(self._cb_mirror, HABApp.openhab.events.ItemStateChangedEventFilter())
        self._config.items.sleeping_state.listen_event(self._cb_sleeping_state, HABApp.openhab.events.ItemStateChangedEventFilter())
        self._config.items.presence_state.listen_event(self._cb_presence_state, HABApp.openhab.events.ItemStateChangedEventFilter())

    def _get_initial_state(self, default_value: str = "initial") -> str:  # noqa: ARG002
        """Get initial state of state machine.

        Args:
            default_value: default / initial state

        Returns:
            if OpenHAB item has a state it will return it, otherwise return the given default value
        """
        return "Manual" if self._config.items.manual.is_on() else "Auto"

    def on_enter_Auto_Init(self) -> None:  # noqa: N802
        """Callback, which is called on enter of Auto_Init state."""
        if self._config.items.light_main.is_on():
            self.to_Auto_On()
        else:
            self.to_Auto_Off()

    def on_enter_Auto_On_Init(self) -> None:  # noqa: N802
        """Callback, which is called on enter of Auto_On_Init state."""
        if self._mirror_is_on():
            self.to_Auto_On_MainAndMirror()
        elif self._is_day():
            self.to_Auto_On_MainDay()
        else:
            self.to_Auto_On_MainNight()

    def _update_openhab_state(self) -> None:
        """Update OpenHAB state item and other states.

        This should method should be set to "after_state_change" of the state machine.
        """
        if self.state != self._previous_state:
            super()._update_openhab_state()
            self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

            self._set_outputs()
            self._previous_state = self.state

    def _set_outputs(self) -> None:
        """Set outputs to OpenHAB items."""
        if self.state == "Manual":
            self._switch_on_via_increase = False
            return

        if self.state == "Auto_Off":
            habapp_rules.core.helper.send_if_different(self._config.items.light_mirror, "OFF")
            if self._light_main_observer.value:
                self._light_main_observer.send_command(0)
        elif self.state == "Auto_On_MainDay":
            habapp_rules.core.helper.send_if_different(self._config.items.light_main_hcl, "ON")
        elif self.state == "Auto_On_MainNight":
            if not self._switch_on_via_increase:
                extended_sleep_brightness = self._config.parameter.brightness_night_extended or self._config.parameter.brightness_night
                target_brightness = extended_sleep_brightness if self._is_extended_sleep() else self._config.parameter.brightness_night
                self._light_main_observer.send_command(target_brightness)
            habapp_rules.core.helper.send_if_different(self._config.items.light_main_color, self._config.parameter.color_night)
        elif self.state == "Auto_On_MainAndMirror":
            habapp_rules.core.helper.send_if_different(self._config.items.light_main_color, self._config.parameter.color_mirror_sync)
            new_brightness = max(self._config.parameter.min_brightness_mirror_sync, self._light_main_observer.value)
            self._light_main_observer.send_command(new_brightness)

        self._switch_on_via_increase = False

    def _is_day(self) -> bool:
        """Check if it is day.

        Returns:
            True if it is day
        """
        if self._config.items.sleeping_state.value != habapp_rules.system.SleepState.AWAKE.value:
            return False

        return time.time() - self._sleep_end_time > self._config.parameter.extended_sleep_time

    def _is_extended_sleep(self) -> bool:
        """Check if it is extended sleep.

        Returns:
            True if it is extended sleep
        """
        return time.time() - self._sleep_end_time <= self._config.parameter.extended_sleep_time

    def _mirror_is_on(self) -> bool:
        """Check if mirror light is on.

        Returns:
            True if mirror light is on
        """
        return self._config.items.light_mirror.is_on()

    def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if the 'manual' item changed.

        Args:
            event: event, which triggered this callback
        """
        if event.value == "ON":
            self.manual_on()
        else:
            self.manual_off()

    def _cb_hand_on(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        """Callback which is triggered if a hand action was detected.

        Args:
            event: event, which triggered this callback
        """
        self._switch_on_via_increase = event.value == "INCREASE"
        self.hand_on()

    def _cb_hand_off(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        self.hand_off()

    def _cb_mirror(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:  # noqa: ARG002
        if self._config.items.light_mirror.is_on():
            self.mirror_on()
        else:
            self.mirror_off()

    def _cb_sleeping_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
            self.sleep_started()

        if event.value == habapp_rules.system.SleepState.AWAKE.value:
            self._sleep_end_time = time.time()

    def _cb_presence_state(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
        if event.value == habapp_rules.system.PresenceState.LEAVING.value:
            self.leaving()



"""Test Presence rule."""

import collections
import unittest
import unittest.mock

import HABApp.rule.rule

import habapp_rules.actors.state_observer
import habapp_rules.core.state_machine_rule
import tests.helper.oh_item
import tests.helper.test_case_base
import tests.helper.timer


class TestStateObserverSwitch(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing StateObserver for switch item."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch", None)

        self._cb_on = unittest.mock.MagicMock()
        self._cb_off = unittest.mock.MagicMock()
        self._observer_switch = habapp_rules.actors.state_observer.StateObserverSwitch("Unittest_Switch", cb_on=self._cb_on, cb_off=self._cb_off)

    def test_command_from_habapp(self) -> None:
        """Test HABApp rule triggers a command -> no manual should be detected."""
        for value in ["OFF", "OFF", "ON", "ON", "OFF"]:
            self._observer_switch.send_command(value)
            tests.helper.oh_item.item_command_event("Unittest_Switch", value)
            tests.helper.oh_item.item_state_change_event("Unittest_Switch", value)
            self._cb_on.assert_not_called()
            self._cb_off.assert_not_called()

    def test_manu_from_openhab(self) -> None:
        """Test manual detection from openHAB."""
        TestCase = collections.namedtuple("TestCase", "command, cb_on_called, cb_off_called")

        test_cases = [
            TestCase("ON", True, False),
            TestCase("ON", False, False),
            TestCase("OFF", False, True),
            TestCase("OFF", False, False),
            TestCase("ON", True, False),
        ]

        for test_case in test_cases:
            self._cb_on.reset_mock()
            self._cb_off.reset_mock()

            tests.helper.oh_item.item_state_change_event("Unittest_Switch", test_case.command)

            self.assertEqual(test_case.cb_on_called, self._cb_on.called)
            self.assertEqual(test_case.cb_off_called, self._cb_off.called)

            if test_case.cb_on_called:
                self._cb_on.assert_called_with(unittest.mock.ANY)
            if test_case.cb_off_called:
                self._cb_off.assert_called_with(unittest.mock.ANY)

    def test_manu_from_extern(self) -> None:
        """Test manual detection from extern."""
        TestCase = collections.namedtuple("TestCase", "command, cb_on_called, cb_off_called")

        test_cases = [
            TestCase("ON", True, False),
            TestCase("ON", False, False),
            TestCase("OFF", False, True),
            TestCase("OFF", False, False),
            TestCase("ON", True, False),
        ]

        for test_case in test_cases:
            self._cb_on.reset_mock()
            self._cb_off.reset_mock()

            tests.helper.oh_item.item_state_change_event("Unittest_Switch", test_case.command)

            self.assertEqual(test_case.cb_on_called, self._cb_on.called)
            self.assertEqual(test_case.cb_off_called, self._cb_off.called)
            if test_case.cb_on_called:
                self._cb_on.assert_called_with(unittest.mock.ANY)
            if test_case.cb_off_called:
                self._cb_off.assert_called_with(unittest.mock.ANY)
            tests.helper.oh_item.item_state_change_event("Unittest_Switch", test_case.command)
            self.assertEqual(test_case.command == "ON", self._observer_switch.value)

    def test_send_command_exception(self) -> None:
        """Test if correct exceptions is raised."""
        with self.assertRaises(ValueError):
            self._observer_switch.send_command(2)


class TestStateObserverDimmer(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing StateObserver for dimmer item."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer_ctr", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.SwitchItem, "Unittest_Switch_ctr", None)

        self._cb_on = unittest.mock.MagicMock()
        self._cb_off = unittest.mock.MagicMock()
        self._cb_changed = unittest.mock.MagicMock()
        self._observer_dimmer = habapp_rules.actors.state_observer.StateObserverDimmer("Unittest_Dimmer", cb_on=self._cb_on, cb_off=self._cb_off, cb_brightness_change=self._cb_changed, control_names=["Unittest_Dimmer_ctr"])

    def test_init(self) -> None:
        """Test init of StateObserverDimmer."""
        self.assertEqual([], self._observer_dimmer._StateObserverBase__group_items)
        self.assertEqual(1, len(self._observer_dimmer._StateObserverBase__control_items))
        self.assertEqual("Unittest_Dimmer_ctr", self._observer_dimmer._StateObserverBase__control_items[0].name)

        observer_dimmer = habapp_rules.actors.state_observer.StateObserverDimmer("Unittest_Dimmer", cb_on=self._cb_on, cb_off=self._cb_off, cb_brightness_change=self._cb_changed, group_names=["Unittest_Dimmer_ctr"])
        self.assertEqual(1, len(observer_dimmer._StateObserverBase__group_items))
        self.assertEqual("Unittest_Dimmer_ctr", observer_dimmer._StateObserverBase__group_items[0].name)
        self.assertEqual([], observer_dimmer._StateObserverBase__control_items)

    def test__check_item_types(self) -> None:
        """Test if wrong item types are detected correctly."""
        with self.assertRaises(TypeError) as context:
            habapp_rules.actors.state_observer.StateObserverDimmer("Unittest_Dimmer", cb_on=self._cb_on, cb_off=self._cb_off, control_names=["Unittest_Dimmer_ctr", "Unittest_Switch_ctr"])
        self.assertEqual("Found items with wrong item type. Expected: DimmerItem. Wrong: Unittest_Switch_ctr <SwitchItem>", str(context.exception))

    def test_command_from_habapp(self) -> None:
        """Test HABApp rule triggers a command -> no manual should be detected."""
        for value in [100, 0, 30, 100, 0, "ON", "OFF", 0, 80]:
            self._observer_dimmer.send_command(value)
            tests.helper.oh_item.item_command_event("Unittest_Dimmer", value)
            tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", value)
            self._cb_on.assert_not_called()
            self._cb_off.assert_not_called()
            self._cb_changed.assert_not_called()

    def test_manu_from_ctr(self) -> None:
        """Test manual detection from control item."""
        TestCase = collections.namedtuple("TestCase", "command, state, cb_on_called")

        test_cases = [TestCase("INCREASE", 30, True), TestCase("INCREASE", 40, False), TestCase("DECREASE", 20, False)]

        for test_case in test_cases:
            self._cb_on.reset_mock()
            self._cb_off.reset_mock()

            tests.helper.oh_item.item_command_event("Unittest_Dimmer_ctr", test_case.command)

            # cb_on called
            self.assertEqual(test_case.cb_on_called, self._cb_on.called)
            if test_case.cb_on_called:
                self._cb_on.assert_called_once_with(unittest.mock.ANY)

            # cb_off not called
            self._cb_off.assert_not_called()

            tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", test_case.state)
            self.assertEqual(test_case.state, self._observer_dimmer.value)

    def test_basic_behavior_on_knx(self) -> None:
        """Test basic behavior. Switch ON via KNX."""
        # === Switch ON via KNX button ===
        # set initial state
        self._cb_on.reset_mock()
        self._observer_dimmer._value = 0
        self._observer_dimmer._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 100)
        self.assertEqual(100, self._observer_dimmer.value)
        self._cb_on.assert_called_once_with(unittest.mock.ANY)

        # === Switch ON via KNX value ===
        # set initial state
        self._cb_on.reset_mock()
        self._observer_dimmer._value = 0
        self._observer_dimmer._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 42)
        self.assertEqual(42, self._observer_dimmer.value)
        self._cb_on.assert_called_once_with(unittest.mock.ANY)

        # === Switch ON via KNX from group ===
        # set initial state
        self._cb_on.reset_mock()
        self._observer_dimmer._value = 0
        self._observer_dimmer._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 80)
        self.assertEqual(80, self._observer_dimmer.value)
        self._cb_on.assert_called_once_with(unittest.mock.ANY)

        # === Value via KNX from group ===
        # set initial state
        self._cb_on.reset_mock()
        self._observer_dimmer._value = 0
        self._observer_dimmer._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", 60)
        self.assertEqual(60, self._observer_dimmer.value)
        self._cb_on.assert_called_once_with(unittest.mock.ANY)

    def test_manu_from_openhab(self) -> None:
        """Test manual detection from control item."""
        TestCase = collections.namedtuple("TestCase", "command, state, cb_on_called, cb_off_called")
        self._observer_dimmer._StateObserverDimmer__last_received_value = 0

        test_cases = [
            TestCase(100, 100, True, False),
            TestCase(100, 100, False, False),
            TestCase(0, 0, False, True),
            TestCase("ON", 100.0, True, False),
            TestCase("OFF", 0.0, False, True),
        ]

        for test_case in test_cases:
            self._cb_on.reset_mock()
            self._cb_off.reset_mock()
            tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", test_case.command)

            self.assertEqual(test_case.cb_on_called, self._cb_on.called)
            self.assertEqual(test_case.cb_off_called, self._cb_off.called)
            if test_case.cb_on_called:
                self._cb_on.assert_called_once_with(unittest.mock.ANY)
            if test_case.cb_off_called:
                self._cb_off.assert_called_once_with(unittest.mock.ANY)
            tests.helper.oh_item.item_state_change_event("Unittest_Dimmer", test_case.state)
            self.assertEqual(test_case.state, self._observer_dimmer.value)

    def test_check_manual(self) -> None:
        """Test method _check_manual."""
        TestCase = collections.namedtuple("TestCase", "event, current_value, on_called, off_called, change_called")

        test_cases = [
            TestCase(HABApp.openhab.events.ItemCommandEvent("any", "ON"), 0, True, False, False),
            TestCase(HABApp.openhab.events.ItemCommandEvent("any", "OFF"), 42, False, True, False),
            TestCase(HABApp.openhab.events.ItemCommandEvent("any", 0), 0, False, False, False),
            TestCase(HABApp.openhab.events.ItemCommandEvent("any", 42), 0, True, False, False),
            TestCase(HABApp.openhab.events.ItemCommandEvent("any", 0), 42, False, True, False),
            TestCase(HABApp.openhab.events.ItemCommandEvent("any", 42), 17, False, False, True),
            TestCase(HABApp.openhab.events.ItemCommandEvent("any", 42), 80, False, False, True),
        ]

        with (
            unittest.mock.patch.object(self._observer_dimmer, "_cb_on") as cb_on_mock,
            unittest.mock.patch.object(self._observer_dimmer, "_cb_off") as cb_off_mock,
            unittest.mock.patch.object(self._observer_dimmer, "_cb_brightness_change") as cb_change_mock,
        ):
            for test_case in test_cases:
                cb_on_mock.reset_mock()
                self._observer_dimmer._value = test_case.current_value
                cb_on_mock.reset_mock()
                cb_off_mock.reset_mock()
                cb_change_mock.reset_mock()

                self._observer_dimmer._check_manual(test_case.event)

                self.assertEqual(cb_on_mock.called, test_case.on_called)
                self.assertEqual(cb_off_mock.called, test_case.off_called)
                self.assertEqual(cb_change_mock.called, test_case.change_called)

                if test_case.on_called:
                    cb_on_mock.assert_called_once_with(test_case.event)
                if test_case.off_called:
                    cb_off_mock.assert_called_once_with(test_case.event)
                if test_case.change_called:
                    cb_change_mock.assert_called_once_with(test_case.event)

    def test_cb_group_item(self) -> None:
        """Test _cb_group_item."""
        self._observer_dimmer._group_last_event = 0
        with unittest.mock.patch("time.time") as time_mock, unittest.mock.patch.object(self._observer_dimmer, "_check_manual") as check_manual_mock:
            time_mock.return_value = 10
            self._observer_dimmer._cb_group_item(HABApp.openhab.events.ItemStateUpdatedEvent("item_name", "ON"))
            time_mock.return_value = 10.2
            self._observer_dimmer._cb_group_item(HABApp.openhab.events.ItemStateUpdatedEvent("item_name", "ON"))
        check_manual_mock.assert_called_once()

    def test_send_command_exception(self) -> None:
        """Test if correct exceptions is raised."""
        with self.assertRaises(ValueError):
            self._observer_dimmer.send_command(None)

        with self.assertRaises(ValueError):
            self._observer_dimmer.send_command("dimmer")


class TestStateObserverRollerShutter(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing StateObserver for switch item."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.RollershutterItem, "Unittest_RollerShutter", None)
        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.RollershutterItem, "Unittest_RollerShutter_ctr", None)

        self._cb_manual = unittest.mock.MagicMock()
        self._observer_jalousie = habapp_rules.actors.state_observer.StateObserverRollerShutter("Unittest_RollerShutter", cb_manual=self._cb_manual, control_names=["Unittest_RollerShutter_ctr"])

    def test_init(self) -> None:
        """Test init of StateObserverDimmer."""
        self.assertEqual([], self._observer_jalousie._StateObserverBase__group_items)
        self.assertEqual(1, len(self._observer_jalousie._StateObserverBase__control_items))
        self.assertEqual("Unittest_RollerShutter_ctr", self._observer_jalousie._StateObserverBase__control_items[0].name)

        observer_jalousie = habapp_rules.actors.state_observer.StateObserverRollerShutter("Unittest_RollerShutter", cb_manual=self._cb_manual, group_names=["Unittest_RollerShutter_ctr"])
        self.assertEqual(1, len(observer_jalousie._StateObserverBase__group_items))
        self.assertEqual("Unittest_RollerShutter_ctr", observer_jalousie._StateObserverBase__group_items[0].name)
        self.assertEqual([], observer_jalousie._StateObserverBase__control_items)

    def test_command_from_habapp(self) -> None:
        """Test HABApp rule triggers a command -> no manual should be detected."""
        for value in [100, 0, 30, 100.0, 0.0]:
            self._observer_jalousie.send_command(value)
            tests.helper.oh_item.item_command_event("Unittest_RollerShutter", value)
            tests.helper.oh_item.item_state_change_event("Unittest_RollerShutter", value)
            self._cb_manual.assert_not_called()

    def test_command_from_habapp_exception(self) -> None:
        """Test HABApp rule triggers a command with wrong type."""
        with self.assertRaises(TypeError):
            self._observer_jalousie.send_command("UP")
        self._cb_manual.assert_not_called()

    def test_manu_from_ctr(self) -> None:
        """Test manual detection from control item."""
        TestCase = collections.namedtuple("TestCase", "command, cb_manual_called")

        test_cases = [
            TestCase("DOWN", True),
            TestCase("DOWN", True),
            TestCase("UP", True),
            TestCase("UP", True),
            TestCase(0, False),
            TestCase(100, False),
        ]

        for test_case in test_cases:
            self._cb_manual.reset_mock()

            tests.helper.oh_item.item_command_event("Unittest_RollerShutter_ctr", test_case.command)

            self.assertEqual(test_case.cb_manual_called, self._cb_manual.called)
            if test_case.cb_manual_called:
                self._cb_manual.assert_called_once_with(unittest.mock.ANY)

    def test_basic_behavior_on_knx(self) -> None:
        """Test basic behavior. Switch ON via KNX."""
        # === Switch ON via KNX button ===
        # set initial state
        self._cb_manual.reset_mock()
        self._observer_jalousie._value = 0
        self._observer_jalousie._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_RollerShutter", 100)
        self.assertEqual(100, self._observer_jalousie.value)
        self._cb_manual.assert_called_once_with(unittest.mock.ANY)

        # === Switch ON via KNX value ===
        # set initial state
        self._cb_manual.reset_mock()
        self._observer_jalousie._value = 0
        self._observer_jalousie._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_RollerShutter", 42)
        self.assertEqual(42, self._observer_jalousie.value)
        self._cb_manual.assert_called_once_with(unittest.mock.ANY)

        # === Switch ON via KNX from group ===
        # set initial state
        self._cb_manual.reset_mock()
        self._observer_jalousie._value = 0
        self._observer_jalousie._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_RollerShutter", 80)
        self.assertEqual(80, self._observer_jalousie.value)
        self._cb_manual.assert_called_once_with(unittest.mock.ANY)

        # === Value via KNX from group ===
        # set initial state
        self._cb_manual.reset_mock()
        self._observer_jalousie._value = 0
        self._observer_jalousie._StateObserverDimmer__last_received_value = 0
        # In real system, this command is triggered about 2 sec later
        tests.helper.oh_item.item_state_change_event("Unittest_RollerShutter", 60)
        self.assertEqual(60, self._observer_jalousie.value)
        self._cb_manual.assert_called_once_with(unittest.mock.ANY)

    def test_check_manual(self) -> None:
        """Test _check_manual."""
        TestCase = collections.namedtuple("TestCase", "event, value, tolerance, cb_called")

        test_cases = [
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", None, None), 0, 0, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 0, None), None, 0, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 0, None), 0, 0, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 10, None), 10, 0, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 1, None), 0, 0, True),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 10, None), 0, 0, True),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 1, None), 0, 2, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 2, None), 0, 2, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 3, None), 0, 2, True),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 9, None), 10, 2, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 8, None), 10, 2, False),
            TestCase(HABApp.openhab.events.ItemStateChangedEvent("any", 7, None), 10, 2, True),
        ]

        with unittest.mock.patch.object(self._observer_jalousie, "_trigger_callback") as trigger_callback_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    trigger_callback_mock.reset_mock()
                    self._observer_jalousie._value = test_case.value
                    self._observer_jalousie._value_tolerance = test_case.tolerance

                    self._observer_jalousie._check_manual(test_case.event)

                    self.assertEqual(test_case.cb_called, trigger_callback_mock.called)


class TestStateObserverNumber(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing StateObserver for number item."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.NumberItem, "Unittest_Number", None)

        self._cb_manual = unittest.mock.MagicMock()
        self._observer_number = habapp_rules.actors.state_observer.StateObserverNumber("Unittest_Number", self._cb_manual, value_tolerance=0.1)

    def test_command_from_habapp(self) -> None:
        """Test HABApp rule triggers a command -> no manual should be detected."""
        for value in [0, 0, 42, 42, 0]:
            self._observer_number.send_command(value)
            tests.helper.oh_item.item_command_event("Unittest_Number", value)
            tests.helper.oh_item.item_state_change_event("Unittest_Number", value)
            self._cb_manual.assert_not_called()

    def test_manu_from_openhab(self) -> None:
        """Test manual detection from openHAB."""
        TestCase = collections.namedtuple("TestCase", "command, cb_manual_called")

        test_cases = [
            TestCase(0, False),
            TestCase(42, True),
            TestCase(0, True),
            TestCase(0, False),
            TestCase(1000, True),
        ]

        for test_case in test_cases:
            self._cb_manual.reset_mock()

            tests.helper.oh_item.item_state_change_event("Unittest_Number", test_case.command)

            self.assertEqual(test_case.cb_manual_called, self._cb_manual.called)
            if test_case.cb_manual_called:
                self._cb_manual.assert_called_with(unittest.mock.ANY)

    def test_manu_from_extern(self) -> None:
        """Test manual detection from extern."""
        TestCase = collections.namedtuple("TestCase", "command, cb_manual_called")

        test_cases = [
            TestCase(0, False),
            TestCase(42, True),
            TestCase(0, True),
            TestCase(0, False),
            TestCase(1000, True),
        ]

        for test_case in test_cases:
            self._cb_manual.reset_mock()

            tests.helper.oh_item.item_state_change_event("Unittest_Number", test_case.command)

            self.assertEqual(test_case.cb_manual_called, self._cb_manual.called)
            if test_case.cb_manual_called:
                self._cb_manual.assert_called_with(unittest.mock.ANY)
            tests.helper.oh_item.item_state_change_event("Unittest_Number", test_case.command)
            self.assertEqual(test_case.command, self._observer_number.value)

    def test_check_manual(self) -> None:
        """Test _check_manual."""
        TestCase = collections.namedtuple("TestCase", "last_value, new_value, manual_expected")

        test_cases = [
            # same value -> False
            TestCase(0, 0, False),
            TestCase(1, 1, False),
            TestCase(100, 100, False),
            # diff < 0.1 -> False
            TestCase(1, 0.9, False),
            TestCase(1, 1.09, False),
            TestCase(0.9, 1, False),
            TestCase(1.09, 1, False),
            TestCase(0, -0.1, False),
            TestCase(0, 0.1, False),
            TestCase(-0.1, 0, False),
            TestCase(0.1, 0, False),
            # diff > 0.1 -> True
            TestCase(0, 0.2, True),
            TestCase(0, -0.2, True),
            TestCase(0.2, 0, True),
            TestCase(-0.2, 0, True),
        ]

        with unittest.mock.patch.object(self._observer_number, "_trigger_callback") as trigger_cb_mock:
            for test_case in test_cases:
                with self.subTest(test_case=test_case):
                    trigger_cb_mock.reset_mock()
                    self._observer_number._value = test_case.last_value
                    self._observer_number._check_manual(HABApp.openhab.events.ItemStateChangedEvent("some_name", test_case.new_value, None))
                    if test_case.manual_expected:
                        trigger_cb_mock.assert_called_once()
                    else:
                        trigger_cb_mock.assert_not_called()

    def test_send_command_exception(self) -> None:
        """Test if correct exceptions is raised."""
        with self.assertRaises(TypeError):
            self._observer_number.send_command("ON")


class TestStateObserverSlat(tests.helper.test_case_base.TestCaseBase):
    """Tests cases for testing StateObserver for number / dimmer item used as slat item."""

    def setUp(self) -> None:
        """Setup test case."""
        tests.helper.test_case_base.TestCaseBase.setUp(self)

        tests.helper.oh_item.add_mock_item(HABApp.openhab.items.DimmerItem, "Unittest_Dimmer", None)
        self._cb_manual = unittest.mock.MagicMock()
        self._observer_slat = habapp_rules.actors.state_observer.StateObserverSlat("Unittest_Dimmer", self._cb_manual)

    def test_check_manual(self) -> None:
        """Test _check_manual."""
        # value is 0
        with unittest.mock.patch("threading.Timer") as timer_mock, unittest.mock.patch("habapp_rules.actors.state_observer.StateObserverNumber._check_manual") as base_check_manual_mock:
            self._observer_slat._check_manual(event := HABApp.openhab.events.ItemStateChangedEvent("any", 0, 42))
        timer_mock.assert_called_once_with(3, self._observer_slat._StateObserverSlat__cb_check_manual_delayed, [event])
        base_check_manual_mock.assert_not_called()

        # value is 100
        with unittest.mock.patch("threading.Timer") as timer_mock, unittest.mock.patch("habapp_rules.actors.state_observer.StateObserverNumber._check_manual") as base_check_manual_mock:
            self._observer_slat._check_manual(event := HABApp.openhab.events.ItemStateChangedEvent("any", 100, 42))
        timer_mock.assert_called_once_with(3, self._observer_slat._StateObserverSlat__cb_check_manual_delayed, [event])
        base_check_manual_mock.assert_not_called()

        # other value | timer not running
        with unittest.mock.patch("threading.Timer") as timer_mock, unittest.mock.patch("habapp_rules.actors.state_observer.StateObserverNumber._check_manual") as base_check_manual_mock:
            self._observer_slat._check_manual(event := HABApp.openhab.events.ItemStateChangedEvent("any", 80, 42))
        timer_mock.assert_not_called()
        base_check_manual_mock.assert_called_once_with(self._observer_slat, event)

    def test__cb_check_manual_delayed(self) -> None:
        """Test __cb_check_manual_delayed."""
        event_mock = unittest.mock.MagicMock()
        with unittest.mock.patch("habapp_rules.actors.state_observer.StateObserverNumber._check_manual") as base_check_manual_mock:
            self._observer_slat._StateObserverSlat__cb_check_manual_delayed(event_mock)
        base_check_manual_mock.assert_called_once_with(self._observer_slat, event_mock)

    def test_on_rule_removed(self) -> None:
        """Test on_rule_removed."""
        with unittest.mock.patch.object(self._observer_slat, "_stop_timer_manual") as stop_timer_mock:
            self._observer_slat.on_rule_removed()
        stop_timer_mock.assert_called_once()
